# -*- coding: utf-8 -*-
"""
qspec.stats
===========

Module for doing statistical analysis.
"""

import numpy as np
import scipy.stats as st
import scipy.integrate as si
from scipy.optimize import root, minimize
import matplotlib.pyplot as plt

from qspec.qtypes import *
from qspec import tools
from qspec.analyze import curve_fit

__all__ = ['Observable', 'add', 'mul', 'average', 'median', 'mode_lognormal', 'estimate_skewnorm', 'propagate',
           'propagate_fit', 'combined_pdf', 'relevant_interval', 'uniform', 'uniform_pumped', 'info']


class Observable(float):
    """
    A float object which has a 'label' and optionally both a left- and right-sided or a symmetric uncertainty.
    """
    def __new__(cls, x: Union[SupportsFloat, SupportsIndex, str, bytes, bytearray],
                std: scalar = None, std_2: scalar = None, label: str = None):
        return super(Observable, cls).__new__(cls, x)

    def __init__(self, x: Union[SupportsFloat, SupportsIndex, str, bytes, bytearray],
                 std: scalar = None, std_2: scalar = None, label: str = None):
        """
        :param x: The value of the observable.
        :param label: The label of the observable.
        :param std: The left-sided 1-sigma percentile relative to 'x' if 'std_2' is specified,
         else the standard deviation of the observable.
        :param std_2: The right-sided 1-sigma percentile relative to 'x' if 'std' is specified,
         else the standard deviation of the observable.
        """
        float.__init__(float(x))
        self.label = label
        self.std = std if std is not None else std_2
        self.std_2 = std_2 if std is not None else std
        self.popt: Union[array_iter, None] = None
        self.est: Union[array_iter, None] = None
        if std_2 is not None:
            try:
                self.est = estimate_skewnorm(float(self), self.std, self.std_2)
            except ValueError as e:
                tools.printw(repr(e))
                tools.printw('The real PDF might not be sufficiently described by a skew normal distribution.')

    def __repr__(self):
        return 'Observable({}, {}, {}, \'{}\')'.format(super().__repr__(), self.std, self.std_2, self.label)

    def __str__(self):
        if self.std is None:
            return '{}'.format(super().__repr__())
        elif self.std_2 is None:
            return '{}(+-{})'.format(super().__repr__(), self.std)
        else:
            return '{}(-{}+{})'.format(super().__repr__(), self.std, self.std_2)

    def __add__(self, other):
        return propagate(add, [self, other])

    def __radd__(self, other):
        return propagate(add, [other, self])

    def __mul__(self, other):
        return propagate(mul, [self, other])

    def __rmul__(self, other):
        return propagate(mul, [other, self])

    def __pow__(self, power, modulo=None):
        return propagate(pow, [self, power])

    def __rpow__(self, power, modulo=None):
        return propagate(pow, [power, self])

    def set_popt(self, popt: array_like = None):
        """
        :param popt: The new value of 'popt'.
        :returns: None. Sets the 'popt' attribute.
        """
        if popt is None:
            self.popt = popt
        else:
            self.popt = np.asarray(popt)

    def rvs(self, size: int = 1) -> ndarray:
        """
        :param size: The defining number of random variates (default is 1).
        :returns: Random variates of the observable of given 'size'. If asymmetric uncertainties are specified,
         a skew normal distribution is assumed. However, the ratio between the left- and right-sided uncertainty
         must not exceed 1.5.
        """
        if self.std_2 is None:
            return st.norm.rvs(loc=self, scale=0. if self.std is None else self.std, size=size)
        elif self.popt is not None:
            return st.skewnorm.rvs(*self.popt, size=size)
        elif self.est is not None:
            return st.skewnorm.rvs(*self.est, size=size)
        else:
            return np.full(size, np.nan)

    def pdf(self) -> Callable:
        """
        :returns: The probability distribution function (pdf) of the observable.
        """
        if self.std_2 is None:
            return np.zeros_like if self.std is None else st.norm(loc=self, scale=self.std).pdf
        elif self.popt is not None:
            return st.skewnorm(*self.popt).pdf
        elif self.est is not None:
            return st.skewnorm(*self.est).pdf
        else:
            return lambda x: np.full_like(x, np.nan)

    def hist(self, size: int = 1000000, n_bins: int = 200):
        """
        :param size: The defining number of random variates (default is 1,000,000).
        :param n_bins: The number of bins.
        :returns: None. Plots a histogram of the observable.
        """
        y = self.rvs(size)
        n, bins, _ = plt.hist(y, bins=n_bins, density=True, label='Sample data', facecolor='lightgray')
        bins = bins[:-1] + 0.5 * (bins[-1] - bins[0]) / n_bins
        plt.plot(bins, self.pdf()(bins), 'C0', label='Est. PDF')
        plt.legend(loc=1)
        plt.show()


def add(a, b):
    """
    :param a: The first summand.
    :param b: The second summand.
    :returns: The sum of the two summands.
    """
    return a + b


def mul(a, b):
    """
    :param a: The first factor.
    :param b: The second factor.
    :returns: The product of the two factors.
    """
    return a * b


def average(a: array_like, std: array_like = None, cov: array_iter = None, axis: int = None) -> (ndarray, ndarray):
    """
    :param a: The sample data.
    :param axis: The axis along which the average is computed.
    :param std: An array of standard deviations associated with the values in 'a'.
     If specified, the weighted average of the uncorrelated sample data and its standard error is returned.
    :param cov: The covariance matrix associated with the values in 'a'.
     If specified, it overrides 'std' and the weighted average of the correlated sample data
     and its standard error is returned. If no axis is specified, 'cov' must have shape (a.size, a.size)
     with elements associated with the values in the flattened 'a',
     otherwise cov must have either shape (a.shape[axis], a.shape[axis])
     or (..., a.shape[axis - 1], a.shape[axis], a.shape[axis], a.shape[axis + 1], ...).
    :returns: The average and its standard error for a given sample 'a' along the specified 'axis'.
    """
    a = np.asarray(a, dtype=float)
    if std is None and cov is None:  # The average and its standard error.
        av = np.average(a, axis=axis)
        av_d = np.std(a, axis=axis, ddof=1)
        if axis is None:  # The shape of 'a' is ignored ('a' is flattened).
            av_d /= np.sqrt(a.size)
        else:
            av_d /= np.sqrt(a.shape[axis])
    elif std is not None and cov is None:  # The weighted average of uncorrelated data and its standard error.
        std = np.asarray(std, dtype=float)
        av, sum_of_weights = np.average(a, axis=axis, weights=1. / (std ** 2), returned=True)
        av_d = np.sqrt(1. / sum_of_weights)
    else:  # The weighted average of correlated data and its standard error.
        cov = np.asarray(cov, dtype=float)
        if axis is None:  # The shape of 'a' is ignored ('a' is flattened).
            if cov.shape != (a.size, a.size):
                raise ValueError('Shape mismatch between \'a\' {} and \'cov\' {}.'
                                 .format(a.shape, cov.shape))
            av, sum_of_weights = np.average(a, weights=1. / np.diag(cov), returned=True)
            av_d = np.sum([1. / cov[i, i] + 2. * np.sum([cov[i, j] / (cov[i, i] * cov[j, j])
                                                         for j in range(a.size) if i < j])
                           for i in range(a.size)])
        else:
            if a.size != a.shape[0] and cov.shape == (a.size, a.size):
                raise ValueError('Shape mismatch between \'a\' {} and \'cov\' {} for specified axis {}.'
                                 .format(a.shape, cov.shape, axis))
            elif cov.shape == (a.shape[axis], a.shape[axis]):
                av, sum_of_weights = np.average(a, axis=axis, weights=1. / np.diag(cov), returned=True)
                av_d = np.sum([1. / cov[i, i] + 2. * np.sum([cov[i, j] / (cov[i, i] * cov[j, j])
                                                             for j in range(a.shape[axis]) if i < j])
                               for i in range(a.shape[axis])])
            else:
                var = np.diagonal(cov, axis1=axis, axis2=axis + 1)
                axes = [ax for ax in range(len(var.shape))]
                i = axes.pop(-1)
                axes.insert(axis, i)
                var = np.transpose(var, axes=axes)
                sum_of_weights = np.sum(1. / var, axis=axis)
                av = np.sum(a / var, axis=axis) / sum_of_weights
                av_d = np.sum(cov / (np.expand_dims(var, axis=axis + 1) * np.expand_dims(var, axis=axis)),
                              axis=(axis + 1, axis))
        av_d = np.sqrt(av_d) / sum_of_weights
    return av, av_d


def median(a: array_like, axis: int = None) -> (ndarray, ndarray, ndarray):
    """
    :param a: The sample data.
    :param axis: The axis along which the three percentiles are computed.
    :returns: The median (0.5-percentile) as well as the left- (~0.1587) and right-sided (~0.8413) 1-sigma percentile
     of a given sample 'a' along the specified 'axis'.
    """
    med = np.median(a, axis=axis)
    neg = np.percentile(a, 15.8655254, axis=axis)
    pos = np.percentile(a, 84.1344746, axis=axis)
    return med, neg, pos


def estimate_skewnorm(med: scalar, per_0: scalar, per_1: scalar):
    """
    :param med: The median (0.5-percentile) of a random variable.
    :param per_0: The left-sided 1-sigma percentile (~0.1587-percentile) relative to 'med'.
    :param per_1: The right-sided 1-sigma percentile (~0.8413-percentile) relative to 'med'.
    :returns: A size-3-array of the estimated parameters (alpha, mean, std.)
     of a skew normal distribution that matches the given percentiles.
    :raises ValueError: If the ratio between the left-(right-) and right-(left-)sided uncertainty exceeds 1.5.
    """
    if per_0 == 0 or per_1 == 0:
        return None
    if per_0 / per_1 > 1.5 or per_1 / per_0 > 1.5:
        raise ValueError('The ratio between the left-(right-) and right-(left-)sided uncertainty must not exceed 1.5.')
    per = np.array([med, med - per_0, med + per_1])
    y = np.array([0.5, 0.158655254, 0.841344746])

    def f(x):
        return st.skewnorm.cdf(per, *x) - y

    def df(x):
        xi0 = -np.exp(-0.5 * ((per - x[1]) / x[2]) ** 2 * (1. + x[0] ** 2)) / (np.pi * (1. + x[0] ** 2))
        xi1 = -st.skewnorm.pdf(per, *x)
        xi2 = -st.skewnorm.pdf(per, *x) * (per - x[1]) / x[2]
        return np.array([xi0, xi1, xi2]).T

    return root(f, x0=np.array([0., med, per_0]), jac=df).x


def lognormal(x, s=0., loc=0., scale=1., a=1.):
    return a * st.lognorm.pdf(x, s=s, loc=loc, scale=scale)


def mode_lognormal(x, bins=100):
    y, edges = np.histogram(x, bins=bins)
    delta = np.mean(edges[1:] - edges[:-1])
    x = edges[:-1] + 0.5 * delta

    loc = np.min(x) - 0.1 * (np.max(x) - np.min(x))
    scale = np.mean(x - loc)
    s = np.std(np.log(x - loc))
    a = scale * np.max(y)
    p0 = [s, loc, scale, a]

    popt, pcov = curve_fit(lognormal, x, y, p0=p0)
    mode = popt[1] + popt[2] * np.exp(-popt[0] ** 2)
    mode = p0[1] + p0[2] * np.exp(-p0[0] ** 2)

    plt.bar(x, y, width=delta)
    plt.plot(x, lognormal(x, *p0), '-C1')
    plt.plot(x, lognormal(x, *popt), '-C2')
    plt.vlines(mode, 0, np.max(y), colors='k')
    plt.show()
    return mode


def propagate(f: Callable, x: array_like, x_d: array_like = None, cov: array_iter = None,
              unc_places: int = None, sample_size: int = 1000000, rtol: float = 1e-3, atol: float = None,
              force_sym: bool = False, full_output: bool = False, show: bool = False) -> (Observable, list, ndarray):
    """
    :param f: The function to compute. 'f' needs to be vectorized.
    :param x: The input values. If 'x_d' is None, the sample data will be generated with the 'Observable.rvs' function
     which considers asymmetric uncertainties. If an element is not an 'Observable', its uncertainty is assumed to be 0.
    :param x_d: The uncertainties of the input values.
    :param cov: The covariance matrix of the x values. If not None, 'x' are assumed to be distributed according to
     a multivariate normal distribution with covariance 'cov'.
    :param unc_places: The number of significant decimal places the result will be rounded to.
     If None, the result is not rounded.
    :param sample_size: The number of random variates used for the calculation. The default is 1,000,000.
    :param rtol: The relative tolerance, with respect to the median of the resulting sample,
     with which the left- and right-sided uncertainties can deviate before asymmetric uncertainties are used.
    :param atol: The absolute tolerance with which the left- and right-sided uncertainties can deviate,
     before asymmetric uncertainties are used. Overrides 'rtol'.
    :param force_sym: Whether to force symmetric uncertainties. If so, a normal distribution is assumed.
    :param full_output: Whether to return the randomly generated data samples.
    :param show: Whether to show a histogram and estimated PDFs of the computed sample data.
    :returns: An 'Observable' whose uncertainties result from the propagation of the uncertainties
     of the input values 'x' by function 'f'.
     If the uncertainties are asymmetric, the parameters of a skew normal distribution
     are estimated using least-square fitting and are stored to the observable. The value and the two uncertainties are
     the median and the left- (~0.1587) and right-sided (~0.8413) 1-sigma percentiles
     relative to the median, respectively. If the uncertainties are symmetric the observable
     is assumed to be normally distributed. The value and the single uncertainty is then calculated using
     the mean and the standard deviation of the sampled data. If 'full_output' is True, a list of the input samples
     as well as the output sample are returned along with the observable.
    """
    label = f.__name__
    n_bins = 200
    if cov is None:
        if x_d is None:
            rand_x = [x_i.rvs(sample_size) if isinstance(x_i, Observable) else x_i for x_i in x]
        else:
            x = np.asarray(x, dtype=float)
            x_d = np.asarray(x_d, dtype=float)
            if x.size != x_d.size:
                raise ValueError('x and x_d must have the same size but have sizes {} and {}.'.format(x.size, x_d.size))
            rand_x = [x_i if x_d_i == 0. else st.norm.rvs(loc=x_i, scale=x_d_i, size=sample_size)
                      for x_i, x_d_i in zip(x, x_d)]
    else:
        cov = np.asarray(cov)
        rand_x = st.multivariate_normal.rvs(mean=[float(x_i) for x_i in x], cov=cov, size=sample_size).T
    rand_y = np.asarray(f(*rand_x))

    med, per_0, per_1 = median(rand_y)
    mean = np.average(rand_y)
    std = np.std(rand_y, ddof=1)

    plt.figure(num=1, dpi=96)
    n, bins, _ = plt.hist(rand_y, bins=n_bins, density=True, label='Sample data', facecolor='lightgray')
    bins = bins[:-1] + 0.5 * (bins[-1] - bins[0]) / n_bins

    if force_sym or (atol is None and abs(per_1 - per_0) / med <= rtol) \
            or (atol is not None and abs(per_1 - per_0) <= atol):
        if unc_places is not None:
            std, dec = tools.round_to_n(std, unc_places)
            mean = np.around(mean, decimals=dec)
        ob = Observable(mean, label=label, std=float(std))
    else:
        if unc_places is not None:
            per_0, dec_0 = tools.round_to_n(per_0, unc_places)
            per_1, dec_1 = tools.round_to_n(per_1, unc_places)
            dec = np.max([dec_0, dec_1])
            med = np.around(med, decimals=dec)
        ob = Observable(med, label=label, std=float(med - per_0), std_2=float(per_1 - med))
        try:
            popt, _ = curve_fit(st.skewnorm.pdf, bins, n, p0=[0., mean, std])
            ob.set_popt(popt)
        except RuntimeError:
            print('Optimal parameters for a skew normal distribution could not be found through fitting.'
                  ' Using an estimation if possible.')
    
    if show:
        print('Median(sample data): {} (-{} +{})'.format(med, med - per_0, per_1 - med))
        plt.plot(bins, st.norm.pdf(bins, loc=mean, scale=std), 'C0', label='Est. normal')
        if ob.popt is not None:
            med, per_0, per_1 = median(st.skewnorm.rvs(*ob.popt, size=sample_size))
            print('Median(Fit. skew normal): {} (-{} +{})'.format(med, med - per_0, per_1 - med))
            plt.plot(bins, st.skewnorm.pdf(bins, *ob.popt), 'C1', label='Fit. skew normal')
        if ob.est is not None:
            med, per_0, per_1 = median(st.skewnorm.rvs(*ob.est, size=sample_size))
            print('Median(Est. skew normal): {} (-{} +{})'.format(med, med - per_0, per_1 - med))
            plt.plot(bins, st.skewnorm.pdf(bins, *ob.est), 'C2', label='Est. skew normal')
        plt.legend(loc=1)
        plt.show()
    plt.close()

    if full_output:
        return ob, rand_x, rand_y
    else:
        return ob


def propagate_fit(f: Callable, x: array_like, popt: array_like, pcov: array_like, sample_size: int = 1000000):
    """
    :param f: The function to compute. $f$ needs to be vectorized.
    :param x: The input values.
    :param popt: An array of the fitted parameters (compare curve_fit).
    :param pcov: A 2d-array of the estimated covariances (compare curve_fit).
    :param sample_size: The number of random variates used for the calculation. The default is 1,000,000.
    :returns: The median and the $1\\sigma$ percentiles of the sampled function $f$.
    """
    x = np.asarray(x)
    _x = np.expand_dims(x, axis=-1)
    _popt = st.multivariate_normal.rvs(mean=popt, cov=pcov, size=sample_size).T
    _popt = [np.expand_dims(p, axis=tuple(range(len(x.shape)))) for p in _popt]
    med, per_0, per_1 = median(f(_x, *_popt), axis=-1)
    return med, per_0, per_1


def combined_pdf(z: array_like, pdf_1: Callable = st.norm.pdf, pdf_2: Callable = st.norm.pdf,
                 loc_1: float = 0., scale_1: float = 1., loc_2: float = 0., scale_2: float = 1.,
                 operator: str = '+', n: int = 10) -> ndarray:
    """
    :param z: The quantiles of the combined probability density function (pdf).
    :param pdf_1: The pdf of the first random variate.
    :param pdf_2: The pdf of the second random variate.
    :param loc_1: The 'loc' parameter of the first pdf.
    :param scale_1: The 'scale' parameter of the first pdf.
    :param loc_2: The 'loc' parameter of the second pdf.
    :param scale_2: The 'scale' parameter of the second pdf.
    :param operator: The operator that defines the new random variate given by Z = X <operator> Y.
     Currently supported operators are {'+', '*'}.
    :param n: The precision of the numerical integration.
     The integration uses 2 ** n intervals to evaluate the integral.
    :returns: The value of the pdf at the given 'z' quantiles.
    """
    arg = np.asarray(z)
    if arg.shape == ():
        arg = np.expand_dims(arg, axis=0)
    arg = np.expand_dims(arg, axis=-1)

    if operator == '*':
        left = loc_1 * loc_2 - 5. * (loc_1 * loc_2 - (loc_1 - scale_1) * (loc_2 - scale_2))
        right = loc_1 * loc_2 + 5. * (loc_1 * loc_2 - (loc_1 - scale_1) * (loc_2 - scale_2))

        def kernel(x):
            return pdf_1(x, loc=loc_1, scale=scale_1) * pdf_2(arg / x, loc=loc_2, scale=scale_2) / abs(x)

    elif operator == '+':
        left = loc_1 + loc_2 - 10. * np.sqrt(scale_1 ** 2 + scale_2 ** 2)
        right = loc_1 + loc_2 + 10. * np.sqrt(scale_1 ** 2 + scale_2 ** 2)

        def kernel(x):
            return pdf_1(x, loc=loc_1, scale=scale_1) * pdf_2(arg - x, loc=loc_2, scale=scale_2)

    else:
        raise ValueError('Operator \'{}\' not supported.'.format(operator))

    x_range = np.linspace(left, right, 2 ** n + 1)
    dx = x_range[1] - x_range[0]
    x_range = np.expand_dims(x_range, axis=0)
    y = kernel(x_range)
    result = si.romb(y, dx, axis=1)
    return result


def relevant_interval(dist: Callable, *args, show: bool = False, **kwargs):
    """
    :param dist: The probability distribution function (pdf).
    :param args: Additional arguments for the pdf.
    :param show: Whether to show the pdf in the determined interval.
    :param kwargs: Additional keyword arguments for the pdf.
    :returns: An estimation of the interval where most of the probability lies in.
    """
    k = 10
    scale = 1.

    def cost(_x):
        if _x[0] > _x[1]:
            _y = dist(np.linspace(_x[1], _x[0], 2 ** k + 1), *args, **kwargs)
        else:
            _y = dist(np.linspace(_x[0], _x[1], 2 ** k + 1), *args, **kwargs)
        c = (1. - si.romb(_y, abs(_x[1] - _x[0]) / (2 ** k))) * scale
        if show:
            print(_x, c / scale)
        return c

    def d_cost(_x):
        _y = np.array([dist(_x[0], *args, **kwargs), -dist(_x[1], *args, **kwargs)])
        if _x[0] > _x[1]:
            _y = -_y
        return _y * scale

    x0 = np.array([-1., 1.])
    while cost(x0) < 0.1:
        if show:
            print('x0: ', x0)
        x0 *= 0.5
    while cost(x0) > 0.9:
        if show:
            print('x0: ', x0)
        x0 *= 2.
    scale = x0[1]
    ret = minimize(cost, x0, jac=d_cost).x
    if show:
        x = np.linspace(*ret, 2 ** k + 1)
        y = dist(x, *args, **kwargs)
        plt.plot(x, y, label=dist.__name__)
        plt.legend()
        plt.show()
    return ret


""" Probability density functions """


def uniform(x: array_like, width: array_like) -> ndarray:
    """
    :param x: The x quantiles.
    :param width: The width of the uniform distribution.
    :returns: The probability density at 'x'.
    """
    return st.uniform.pdf(x, loc=-0.5 * width, scale=width)


def _uniform_pumped(x, width, gamma_u, depth) -> ndarray:
    """
    :param x: The x quantiles.
    :param width: The width of the uniform distribution.
    :param gamma_u: The fwhm of the pump transition.
    :param depth: The depth of the pump dip in parts of the height of the underlying uniform distribution.
    :returns: A not normalized uniform distribution with a Lorentz shaped dip.
    """
    x = np.asarray(x)
    scalar_true = tools.check_shape((), x, return_mode=True)
    if scalar_true:
        x = np.array([x])
    uni = uniform(x, width)
    nonzero = np.nonzero(uni)
    ret = np.zeros_like(x)
    ret[nonzero] = uni[nonzero] - depth / width * (np.pi * gamma_u / 2.) * st.cauchy.pdf(x[nonzero], scale=gamma_u / 2.)
    if scalar_true:
        return ret[0]
    return ret


def uniform_pumped(x, width, gamma_u, depth) -> ndarray:
    """
    :param x: The x quantiles.
    :param width: The width of the uniform distribution.
    :param gamma_u: The fwhm of the pump transition.
    :param depth: The depth of the pump dip in parts of the height of the underlying uniform distribution.
    :returns: A uniform distribution with a Lorentz-shaped dip.
    """
    _x = np.linspace(-0.5 * width, 0.5 * width, 2 ** 10 + 1)
    _y = _uniform_pumped(_x, width, gamma_u, depth)
    norm = si.romb(_y, width / (2 ** 10))
    return _uniform_pumped(x, width, gamma_u, depth) / norm


def info():
    """
    Plots the 1-sigma percentiles (and their ratio) of a skew normal distribution relative to its median
    for different values of the parameter alpha.
    The plot shows that the ratio between the two 1-sigma percentiles cannot exceed 1.5.

    :returns: None.
    """
    neg = []
    pos = []
    a = np.linspace(-10, 10, 1001)
    for val in a:
        dist = st.skewnorm.rvs(val, loc=0, scale=1, size=100000)
        med = np.median(dist)
        neg.append(med - np.percentile(dist, 15.8655254))
        pos.append(np.percentile(dist, 84.1344746) - med)
    neg, pos = np.array(neg), np.array(pos)
    plt.title('Properties of the skew normal distribution')
    plt.xlabel(r'Parameter $\alpha$')
    plt.ylabel(r'Percentiles ($\sigma$) / Ratio of Percentiles')
    plt.ylim(0., 1.75)
    plt.plot(a, neg, label='left 1-sigma')
    plt.plot(a, pos, label='right 1-sigma')
    plt.plot(a, pos / neg, label='Ratio right/left')
    plt.plot(a, neg / pos, label='Ratio left/right')
    plt.legend(loc='lower center')
    plt.show()



# -*- coding: utf-8 -*-
"""
qspec.tools
===========

Module including mathematical and technical methods.
"""

import os
import sys
import ctypes
from pathlib import Path
import ctypes.wintypes
import shutil
import sqlite3
import numpy as np

from qspec.qtypes import *

__all__ = ['ROMAN_NUMERALS', 'COLORS', 'get_rgb_print_command', 'rgb_to_hex_color', 'print_colored', 'printh', 'printw',
           'printf', 'map_corr_coeff_to_color', 'print_cov', 'get_default_path', 'create_doc_link', 'create_data_dir',
           'get_config_dict', 'fraction', 'check_half_integer', 'half_integer_to_fraction', 'half_integer_to_str',
           'get_val_with_unc', 'roman_to_int', 'odd', 'even', 'get_decimals', 'floor_log2', 'floor_log10', 'round_to_n',
           'factorial', 'asarray_optional', 'in_nested', 'check_iterable', 'make_str_iterable_unique',
           'check_dimension', 'check_shape_like', 'check_shape', 'nan_helper', 'dict_to_list', 'list_to_dict',
           'list_to_excel', 'add_nested_key', 'merge_dicts', 'convolve_dict', 'combine_dicts', 'merge_intervals',
           'absolute', 'absolute_complex', 'angle', 'angle_d', 'transform', 'unit_vector', 'e_r', 'e_theta', 'e_phi',
           'orthonormal_rtp', 'orthonormal', 'rotation_matrix', 'Rotation', 'rotation_to_vector',
           'import_iso_shifts_tilda']


ROMAN_NUMERALS = {'M': 1000, 'D': 500, 'C': 100, 'L': 50, 'X': 10, 'V': 5, 'I': 1}


class COLORS:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    PYPLOT = ['C{}'.format(i) for i in range(10)]


def get_rgb_print_command(r, g, b):
    """
    :param r: The fraction of red (0-255).
    :param g: The fraction of green (0-255).
    :param b: The fraction of blue (0-255).
    :returns: The command str to print rgb colors in the console.
    """
    return '\033[38;2;{};{};{}m'.format(r, g, b)


def rgb_to_hex_color(r, g, b):
    """
    :param r: The fraction of red (0-255).
    :param g: The fraction of green (0-255).
    :param b: The fraction of blue (0-255).
    :returns: A hexadecimal str representation of the `rgb` tuple.
    """
    return '#%02x%02x%02x' % (r, g, b)


def print_colored(specifier, *values, returned=False, **kwargs):
    """
    Print with the specified color.

    :param specifier: str of the color name defined in the COLORS class or an RGB-value (0-255, 0-255, 0-255).
    :param values: The values to print.
    :param returned: Return the str instead of printing it.
    :param kwargs: See print().
    :returns: None.
    """
    if isinstance(specifier, str):
        _c = eval('COLORS.{}'.format(specifier.upper()))
    else:
        _c = get_rgb_print_command(*specifier)
    _values = (_c, ) + values + (COLORS.ENDC, )
    if returned:
        return '{}{}{}'.format(*_values)
    print('{}{}{}'.format(*_values), **kwargs)


def printh(*values, **kwargs):
    """
    Print with the HEADER color (pink).

    :param values: The values to print.
    :param kwargs: See print().
    :returns: None.
    """
    _values = (COLORS.HEADER, ) + values + (COLORS.ENDC, )
    print('{}{}{}'.format(*_values), **kwargs)


def printw(*values, **kwargs):
    """
    Print with the WARNING color (yellow).

    :param values: The values to print.
    :param kwargs: See print().
    :returns: None.
    """
    _values = (COLORS.WARNING, ) + values + (COLORS.ENDC, )
    print('{}{}{}'.format(*_values), **kwargs)


def printf(*values, **kwargs):
    """
    Print with the FAIL color (red).

    :param values: The values to print.
    :param kwargs: See print().
    :returns: None.
    """
    _values = (COLORS.FAIL, ) + values + (COLORS.ENDC, )
    print('{}{}{}'.format(*_values), **kwargs)


def map_corr_coeff_to_color(val, clip=True):
    """

    :param val: A value between -1 and 1.
    :param clip: Whether to clip vals outside -1 and 1. If False, raise ValueError.
    :returns: The RGB values in the range 0-255.
    """
    if clip:
        if np.isnan(val):
            val = 0.
        elif val < -1:
            val = -1
        elif val > 1:
            val = 1
    if np.isnan(val) or val < -1 or val > 1:
        raise ValueError('The correlation coefficient must be in [-1, 1].')
    g = int(round(val * 127 + 127, 0))
    return 255 - g, g, 0

def print_cov(cov, normalize=False, decimals=2):
    """
    Print a covariance "as is" or as color-coded Pearson correlation coefficients.

    :param cov: A covariance matrix.
    :param normalize: Whether to normalize the covariance to be the Pearson correlation coefficient.
    :param decimals: The number of decimal places to be printed.
    :returns: None.
    """
    cov = np.array(cov, dtype=float)
    if normalize:
        norm = np.sqrt(np.diag(cov)[:, None] * np.diag(cov))
        nonzero = norm != 0
        cov[nonzero] /= norm[nonzero]
    cov = np.around(cov + 0., decimals=decimals)
    digits = int(np.floor(np.log10(np.abs(cov.shape[0])))) + 1
    for i, row in enumerate(cov):
        print('{}:   {}'.format(str(i).zfill(digits), '   '.join(
            '{}{}{}'.format(get_rgb_print_command(*map_corr_coeff_to_color(val)),
                            '{:1.2f}'.format(val).rjust(decimals + 3), COLORS.ENDC) for val in row)))


""" System operations """


def get_default_path():
    """
    :returns: The absolute path to the "PyCLS user folder."
    """
    csidl_personal = 5  # My Documents
    shgfp_type_current = 0  # Get current, not default value

    buf = ctypes.create_unicode_buffer(ctypes.wintypes.MAX_PATH)
    ctypes.windll.shell32.SHGetFolderPathW(None, csidl_personal, None, shgfp_type_current, buf)
    return os.path.join(buf.value, 'PyCLS')


def create_doc_link(env_path: str = None):
    """
    :param env_path: The path to the virtual environment where the 'Lib' folder resides in.
     If None, the path is inherited from the currently running Python-executable.
    :returns: None. Creates an html document redirecting to the API-Documentation inside the PyCLS installation.
    """
    if env_path is None:
        env_path = os.path.dirname(sys.executable)
        if os.path.split(env_path)[1].lower() == 'scripts':
            env_path = os.path.split(env_path)[0]
    path = get_default_path()
    html_link = os.path.join(env_path, 'Lib\\site-packages\\PyCLS\\docs\\build\\html_link.txt')
    with open(html_link, 'r') as file:
        s = file.read()
    s = s.replace('ENV-PATH\\Lib', os.path.join(env_path, 'Lib'))
    with open(os.path.join(path, 'API-Documentation.html'), 'w') as file:
        file.write(s)


def create_data_dir(path: str = None, overwrite: bool = False):
    """
    :param path: The path were to create the user folder "PyCLS" (do not include "PyCLS" in 'path').
     If None, the user folder is created in the "Documents" Windows library.
    :param overwrite: Whether to overwrite the current "PyCLS" user folder if it exists.
    :returns: None. Creates the user folder "PyCLS" at the specified 'path'
     or in the "Documents" Windows library if no 'path' is specified.
    """
    path = get_default_path() if path is None else os.path.join(path, 'PyCLS')
    if overwrite or not Path(os.path.join(path, 'PyCLS')).exists():
        Path(path).mkdir(parents=True, exist_ok=True)
        Path(os.path.join(path, 'databases')).mkdir(parents=True, exist_ok=True)

        env_path = os.path.dirname(sys.executable)
        if os.path.split(env_path)[1].lower() == 'scripts':
            env_path = os.path.split(env_path)[0]
        with open(os.path.join(path, 'config.cfg'), 'w') as config:
            config.write('# PyCLS Config file\n\n')
            config.write('Environment path = {}\n'.format(env_path))
            config.write('Data path = {}\n'.format(path))
            config.write('ffmpeg = {}\n'.format('C:\\Program Files\\ffmpeg\\bin\\ffmpeg.exe'))

        src_dir = os.path.join(env_path, 'Lib\\site-packages\\PyCLS\\databases')
        dst_dir = os.path.join(path, 'databases')
        for file in os.listdir(src_dir):
            shutil.copyfile(os.path.join(src_dir, file), os.path.join(dst_dir, file))
        create_doc_link(env_path)
    else:
        raise ValueError('The user folder "...\\Documents\\PyCLS" already exists and \'overwrite\' is set to False. '
                         'If you would like to replace your user folder with the default user folder, '
                         'set \'overwrite=True\'.')


def get_config_dict() -> dict:
    """
    :returns: A dictionary containing the content of the currently used config file.
    """
    path = get_default_path()
    ret = {}
    with open(os.path.join(path, 'config.cfg'), 'r') as config:
        for s in config:
            if '=' in s:
                ret[s[:s.find('=')].strip()] = s[(s.find('=')+1):].strip()
    return ret


""" literal -> numeral operations """


def fraction(r: Union[Rational, str]) -> (int, int):
    """
    :param r: A sympy.Rational or a str with the signature "'num'/'denom'".
    :returns: the numerator and denominator of 'r'.
    """
    if not isinstance(r, Rational) and not isinstance(r, str):
        raise TypeError('Argument must be a sympy.Rational or str, but is {}: {}.'.format(type(r), r))
    string = str(r)
    slash = string.find('/')
    if slash == -1:
        num = int(string)
        denom = 1
    else:
        num = int(string[:slash])
        denom = int(string[(slash+1):])
    return num, denom


def check_half_integer(*args: scalar):
    """
    :param args: Scalar arguments.
    :returns: None. Checks whether the given arguments are multiples of 1/2.
    :raises ValueError: If any argument is not a multiple of 1/2.
    """
    if sum([abs(arg % 0.5) for arg in args]) > 0:
        raise ValueError('The numbers must be multiples of 1/2.')


def half_integer_to_fraction(val: scalar) -> (int, int):
    """
    :param val: A scalar value.
    :returns: The numerator and denominator of the half-integer.
    """
    check_half_integer(val)
    if abs(val % 1) > 0:
        return int(2 * val), 2
    return int(val), 1


def half_integer_to_str(val: scalar, symbol: str = '_'):
    """
    :param val: A scalar value.
    :param symbol: The symbol to use for the fraction.
    :returns: A rational str-representation of a half-integer.
    """
    num, denom = half_integer_to_fraction(val)
    if denom == 1:
        return str(num)
    return '{}{}2'.format(num, symbol)


def get_val_with_unc(val_string: str) -> (float, float):
    """
    :param val_string: The str representation of a number with uncertainty of the format '1.234(56)'.
     Decimal separators can be used or spared arbitrarily.
    :returns: The value and its uncertainty.
    """
    err, decimals = 0., 0
    if val_string == '':
        return err, err
    lbrace = val_string.find('(')
    rbrace = val_string.find(')')
    val = float(val_string) if lbrace == -1 else float(val_string[:lbrace])
    err_string = val_string[(lbrace + 1):rbrace]
    if -1 not in [lbrace, rbrace, len(err_string) - 1]:
        err = float(err_string)
        if '.' in val_string[:lbrace]:
            dot = val_string.find('.')
            decimals = len(val_string[(dot+1):lbrace])
            if '.' not in val_string[lbrace:rbrace]:
                err *= eval('1e-{}'.format(decimals))
            else:
                dot = val_string[lbrace:].find('.')
                decimals = max([decimals, len(val_string[(dot + 1):rbrace])])
    return round(val, decimals), round(err, decimals)


def roman_to_int(roman: str) -> int:
    """
    Convert from Roman numerals to an integer
    [jonrsharpe, https://codereview.stackexchange.com/questions/68297/convert-roman-to-int].

    :param roman: The str representation of a roman number.
    """
    numbers = []
    for char in roman.upper():
        numbers.append(ROMAN_NUMERALS[char])
    total = 0
    num2 = numbers[-1]
    for num1, num2 in zip(numbers, numbers[1:]):
        if num1 >= num2:
            total += num1
        else:
            total -= num1
    return total + num2


""" numeral -> numeral operations """


def odd(x: float) -> int:
    """
    :param x: A float value.
    :returns: The closest odd integer value.
    """
    return int(x) if int(x) % 2 else int(x) + 1


def even(x: float) -> int:
    """
    :param x: A float value.
    :returns: The closest even integer value.
    """
    return int(x) if not int(x) % 2 else int(x) + 1


def get_decimals(x: float) -> int:
    """
    :param x: A float value.
    :return: The number of displayed decimals of the float value 'x'.
    """
    s = str(x)
    dot = s.find('.')
    return len(s[(dot+1):])


def floor_log2(x: array_like) -> ndarray:
    """
    :param x: Scalar values.
    :returns: The closest integer values below the logarithm with basis 10 of the absolute value of 'x'.
    """
    x = np.asarray(x)
    if len(x.shape) > 0:
        nonzero = x.nonzero()
        n = np.zeros_like(x)
        n[nonzero] = np.floor(np.log2(np.abs(x[nonzero]))).astype(int)
        return n
    if x == 0:
        return np.array(0)
    return np.floor(np.log2(np.abs(x))).astype(int)


def floor_log10(x: array_like) -> ndarray:
    """
    :param x: Scalar values.
    :returns: The closest integer values below the logarithm with basis 10 of the absolute value of 'x'.
    """
    x = np.asarray(x)
    if len(x.shape) > 0:
        nonzero = x.nonzero()
        n = np.zeros_like(x)
        n[nonzero] = np.floor(np.log10(np.abs(x[nonzero]))).astype(int)
        return n
    if x == 0:
        return np.array(0)
    return np.floor(np.log10(np.abs(x))).astype(int)


def round_to_n(x: array_like, n: int) -> (scalar, int):
    """
    :param x: The input data.
    :param n: The number of significant decimal places to round to.
    :returns: x rounded to 'n' significant decimal places and
     the corresponding number of decimal places after the decimal point.
    """
    if isinstance(x, ndarray):
        x = x.squeeze().tolist()
        if isinstance(x, list):
            raise TypeError('Only size-1 arrays are supported.')
    if x == 0.:
        return type(x)(0), 0
    else:
        decimals = -int(np.floor(np.log10(np.abs(x)))) + n - 1
        return type(x)(np.around(x, decimals=decimals)), decimals


def factorial(n: array_like):
    """
    :param n: The integer numbers. Float types are cast to int types.
    :returns: n! (array compatible).
    """
    n = np.asarray(n, dtype=int)
    mask = n > 1
    ret = np.ones_like(n)
    if np.any(mask):
        ret[mask] = n[mask] * factorial(n[mask] - 1)
    return ret


""" Iterable operations """


def asarray_optional(a: Optional[array_like], **kwargs):
    """
    :param a: Input data, see numpy docs.
    :param kwargs: The keyword arguments are passed to `numpy.asarray`.
    :returns: None if `a` is None else `numpy.asarray(a, **kwargs)`.
    """
    return None if a is None else np.asarray(a, **kwargs)


def in_nested(a, nested: Iterable) -> bool:
    """
    :param a: The element to look for.
    :param nested: The nested list.
    :returns: Whether `a` is inside the `nested` list.
    """
    if not isinstance(nested, Iterable):
        return False
    if a in nested:
        return True
    ret = False
    for el in nested:
        ret += in_nested(a, el)
        if ret:
            return True
    return ret


def check_iterable(arg: Any, dtype: type = str) -> list:
    """
    :param arg: The argument that is checked.
    :param dtype: The type that has to be matched by the elements of arg or arg itself.
    :returns: 'arg' as a list if it is a 'list', 'tuple' or 'set' of values of type 'dtype',
     a list of the single element 'arg' if it is of type 'dtype' and else an empty list.
    """
    if isinstance(arg, list) or isinstance(arg, tuple) or isinstance(arg, set):
        if all(isinstance(s, dtype) for s in arg):
            return list(arg)
        return []
    elif isinstance(arg, dtype):
        return [arg, ]
    else:
        return []


def make_str_iterable_unique(a: Iterable[str], identifier: Iterable[str] = None):
    """
    :param a: An Iterable of str values.
    :param identifier: An Iterable of str values to append to the values of 'a' if they appear more than once.
     If None, '_i' is used as the identifier for the ith appearance of a value in 'a'.
    :returns: 'a', but the elements which appear more than once are numerated/attached with the identifiers and bunched.
     For example: ['a', 'b', 'c', 'b', 'd'] -> ['a', 'b_0', 'b_1', 'c', 'd'] or
     ['a', 'b', 'c', 'b', 'd'] -> ['a', 'b' + identifier[0], 'b' + identifier[1], 'c', 'd'].
    """
    temp_1 = list(a)
    temp_2 = []
    if identifier is not None:
        identifier = list(identifier)
    for p in temp_1:
        if p not in temp_2:
            temp_2.append(p)
    ret = []
    for p in temp_2:
        m = temp_1.count(p)
        for i in range(m):
            if m == 1:
                ret.append(p)
            else:
                if identifier is None:
                    ret.append(p + '_{}'.format(i))
                else:
                    ret.append(p + '_{}'.format(identifier[i % len(identifier)]))
    return ret


def check_dimension(dim: int, axis: int, *args: ndarray) -> None:
    """
    :param dim: The number of components the axis 'axis' must have.
    :param axis: The axis which must have 'dim' components.
    :param args: The arguments which must have 'dim' components along axis 'axis'.
    :returns: None
    :raises ValueError: All specified arguments must have 'dim' components along axis 'axis'.
    """
    err_list = []
    for arg in args:
        arg = np.asarray(arg)
        if arg.shape[axis] != dim:
            err_list.append(str(arg.shape[axis]))
    if err_list:
        err_msg = 'One argument has {} component(s) instead of {} along axis {}.'\
            .format(', '.join(err_list), dim, axis)
        if len(err_list) > 1:
            err_msg = '{} arguments have {} component(s) instead of {} along axis {}.'\
                .format(len(err_list), ', '.join(err_list), dim, axis)
        raise ValueError(err_msg)


def check_shape_like(*args: ndarray, allow_scalar=True) -> None:
    """
    :param args: The arguments which must have shapes that can be broadcast together.
    :param allow_scalar: Whether scalar values (shape=()) are seen as compatible with arbitrary shapes.
    :returns: None
    :raises ValueError: All specified arguments must have shapes that can be broadcast together.
    """
    err_list = []
    shape = args[-1].shape
    allowed = [len(shape)]
    if allow_scalar:
        allowed.append(0)
    for arg in args[:-1]:
        if len(arg.shape) in allowed:
            if arg.shape != () and any(i != j and i != 1 and j != 1 for i, j in zip(arg.shape, shape)):
                err_list.append(str(arg.shape))
        else:
            err_list.append(str(arg.shape))
    if err_list:
        err_msg = 'Operands could not be broadcast together with shapes {} and {}.'.format(', '.join(err_list), shape)
        raise ValueError(err_msg)


def check_shape(shape: tuple, *args: ndarray, allow_scalar=True, return_mode=False) -> Union[bool, None]:
    """
    :param shape: The shape which must be matched by the specified arguments.
    :param args: The arguments which must match the specified shape.
    :param allow_scalar: Whether scalar values (shape=()) are seen as compatible with arbitrary shapes.
    :param return_mode: Whether to raise a ValueError or return a boolean if any argument does not have shape 'shape'.
    :returns: whether the arguments match the specified shape.
    :raises ValueError: All specified arguments must match the specified shape.
    """
    err_list = []
    allowed = [shape]
    if allow_scalar:
        allowed.append(())
    for arg in args:
        if arg.shape not in allowed:
            err_list.append(str(arg.shape))
    if err_list:
        if return_mode:
            return False
        else:
            err_msg = 'Operand(s) with shape(s) {} do not match shape {}.'.format(', '.join(err_list), shape)
            raise ValueError(err_msg)
    return True


def nan_helper(y: array_like) -> (ndarray, Callable):
    """
    :param y: An array like value.
    :returns: The mask where y is NaN and a function which returns the nonzero indices of an array.
    """
    return np.isnan(y), lambda z: z.nonzero()[0]


def dict_to_list(a: dict) -> (list, list):
    """
    :param a: A dictionary.
    :returns: Two lists with the keys and values of 'a', respectively.
     The lists are in the same order, however, they are not sorted.
    """
    kv = [[k, v] for k, v in a.items()]
    return [k[0] for k in kv], [v[1] for v in kv]


def list_to_dict(values: array_like, keys: array_like = None) -> dict:
    """
    :param values: The values of the dictionary.
    :param keys: The keys of the dictionary. If omitted, the keys will be the indices of the values.
    :returns: A dictionary with the given 'values' and 'keys'.
    """
    v_list = list(np.asarray(values).flatten())
    if keys is None:
        k_list = list(range(len(values)))
    else:
        k_list = list(np.asarray(keys).flatten())
    return {k: v for k, v in zip(k_list, v_list)}


def list_to_excel(*args: array_iter, save: str = None, delimiter: str = '\t', header: str = '', align: str = 'top'):
    """

    :param args: 1- or 2-dimensional iterables to print or save in an excel-compatible format.
    :param save: The filepath, either absolute or relative to the working directory. If None, the output is printed.
    :param delimiter: The delimiter between two columns.
    :param header: Add a header above the content to the output.
    :param align: How to align the columns that are smaller than the largest column.
     Supported alignments are {"top", "bottom", else == "top"}.
    :returns: None. Prints or saves the 'args' in an excel-compatible way.
    """
    columns = []
    s = ''
    for arg in args:
        arg = np.asarray(arg)
        if len(arg.shape) not in [1, 2]:
            raise ValueError('All arguments must be either 1- or 2-dimensional iterables.')
        elif len(arg.shape) == 2:
            for c in arg.T:
                columns.append(list(c))
                s += '{}' + delimiter
        else:
            columns.append(list(arg))
            s += '{}' + delimiter
    if header != '':
        header += '\n'
    c_sizes = [len(c) for c in columns]
    c_max = max(c_sizes)
    rows = []
    if align == 'bottom':
        d = -1
        for i in range(c_max):
            rows.append(s[:-len(delimiter)].format(*['' if i >= n else c[n - i - 1] for c, n in zip(columns, c_sizes)]))
    else:
        d = 1
        for i in range(c_max):
            rows.append(s[:-len(delimiter)].format(*['' if i >= n else c[i] for c, n in zip(columns, c_sizes)]))
    if save is None:
        print(header + '\n'.join(rows[::d]))
        return
    with open(save, mode='w') as file:
        file.write(header)
        file.write('\n'.join(rows[::d]))


def add_nested_key(a: dict, key_list: list, val: array_like) -> dict:
    """
    :param a: The target dictionary.
    :param key_list: A list of nested keys of a.
    :param val: The value that is assigned to the last key in key_list.
    :returns: The dictionary a with val assigned to the last key in key_list.
    """
    if not isinstance(a, dict):
        raise TypeError('Cannot assign another key ({}) since a is not a dictionary.'.format(key_list[0]))
    if len(key_list) == 1:
        a[key_list[0]] = val
    else:
        if key_list[0] not in a.keys():
            a[key_list[0]] = add_nested_key({}, key_list[1:], val)
        else:
            a[key_list[0]] = add_nested_key(a[key_list[0]], key_list[1:], val)
    return a


def merge_dicts(a: dict, b: dict, path=None) -> dict:
    """
    :param a: The first dictionary.
    :param b: The second dictionary.
    :param path: _.
    :returns: b merged into a. Function copied from andrew cooke
     https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries.
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                a[key] = merge_dicts(a[key], b[key], path + [str(key)])
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def convolve_dict(a: dict, *key_lists: Iterable, operator: str = '+', short_keys: bool = True) \
        -> Union[dict, ValueError]:
    """
    :param a: A dictionary to be convolved.
    :param key_lists: An arbitrary number of Iterables of the same length which include keys from 'a'.
    :param operator: The operator which is used to combine the key_lists.
    :param short_keys: Whether to just use the keys of the first key_list
     or representations of the entire operations as the new keys.
    :returns: A dictionary of the shape {key_lists[0] + key_lists[1] + ... :
                                         a[key_lists[0]] + a[key_lists[1]] + ... } if operator == '+'.
    """
    operators = ['+', '-', '*', '/', '%', 'gauss']
    if operator not in operators:
        raise ValueError('Operator \'{}\' not available try one of {}.'.format(operator, operators))
    if operator.lower() == 'gauss':
        return {keys[0] if short_keys else 'Gauss(' + ', '.format(operator).join([str(k) for k in keys]) + ')':
                eval('np.sqrt(' + ' + '.join(['a[{}] ** 2'.format(k) for k in keys]) + ')', {'a': a, 'np': np})
                for keys in zip(*key_lists)}
    return {keys[0] if short_keys else ' {} '.format(operator).join([str(k) for k in keys]):
            eval(' {} '.format(operator).join(['a[{}]'.format(k) for k in keys]), {'a': a})
            for keys in zip(*key_lists)}


def combine_dicts(dicts: list, key_lists: Iterable, operator: str = '+', short_keys: bool = True) \
        -> Union[dict, ValueError]:
    """
    :param dicts: A list of dictionaries to be combined.
    :param key_lists: A list of Iterables of the same lengths which include keys from 'a'.
     Must have the same length as 'dicts'.
    :param operator: The operator which is used to combine the key_lists.
    :param short_keys: Whether to just use the keys of the first key_list
     or representations of the entire operations as the new keys.
    :returns: A dictionary of the shape {key_lists[0] + key_lists[1] + ... :
                                         dicts[0][key_lists[0]] + dicts[1][key_lists[1]] + ... } if operator == '+'.
    """
    operators = ['+', '-', '*', '/', '%', 'gauss']
    if operator not in operators:
        raise ValueError('Operator \'{}\' not available try one of {}.'.format(operator, operators))
    if operator.lower() == 'gauss':
        return {keys[0] if short_keys else 'Gauss(' + ', '.format(operator).join([str(k) for k in keys]) + ')':
                eval('np.sqrt(' + ' + '.join(['dicts[{}][{}] ** 2'.format(i, k)
                                              for i, k in enumerate(keys)]) + ')', {'dicts': dicts, 'np': np})
                for keys in zip(*key_lists)}
    return {keys[0] if short_keys else ' {} '.format(operator).join([str(k) for k in keys]):
            eval(' {} '.format(operator).join(['dicts[{}][{}]'.format(i, k)
                                               for i, k in enumerate(keys)]), {'dicts': dicts})
            for keys in zip(*key_lists)}


def merge_intervals(intervals: array_iter) -> ndarray:
    """
    :param intervals: An iterable of intervals.
     An interval i is itself an iterable of two scalar values. If i[1] < i[0], the interval is reversed.
    :returns: An iterable of non-overlapping intervals.
    """
    inter = np.asarray(intervals)
    check_shape_like(inter, np.zeros((1, 2)))
    inter = np.array([i if i[0] <= i[1] else i[::-1] for i in inter])
    sort = np.argsort(inter[:, 0])
    inter = inter[sort, :]
    new_inter = []
    n = 0
    m = inter.shape[0]
    while n < inter.shape[0] - 1:
        end = inter[n, 1]
        for m, i in enumerate(inter[(n+1):, :]):
            if not end < i[0]:
                end = np.max([end, i[1]])
            else:
                break
        new_inter.append([inter[n, 0], end])
        n += m + 1
    if inter.shape[0] == 1 or inter[-1, 0] > inter[-2, 1]:
        new_inter.append(inter[-1, :])
    new_inter = np.asarray(new_inter)
    return new_inter


def get_subarray(a, i, axis):
    """
    :param a: The array.
    :param i: The indexes to select along the 'axis'
    :param axis: The axis along which the indexes 'i' are selected.
    :returns: The subarray selected with the indexes 'i' along the 'axis'.
    """

    axis += a.ndim if axis < 0 else 0
    index = ((slice(None),) * axis + (i,) + (slice(None),) * (a.ndim - axis - 1))
    return a[index]


""" Vector math """


def absolute(x: array_like, axis=-1) -> ndarray:
    """
    :param x: A real vector or an array of real vectors.
    :param axis: The axis along which the vector components are aligned.
    :returns: The length(s) of the vector(s) x.
    """
    x = np.asarray(x)
    return np.sqrt(np.sum(x ** 2, axis=axis))


def absolute_complex(x: array_like, axis=-1):
    """
    :param x: A complex vector or an array of complex vectors.
    :param axis: The axis along which the vector components are aligned.
    :returns: The length(s) of the vector(s) x.
    """
    x = np.asarray(x)
    return np.sqrt(np.sum(np.abs(x) ** 2, axis=axis))


def angle(x: array_like, y: array_like, axis=-1) -> ndarray:
    """
    :param x: The first vectors (arb. units).
    :param y: The second vectors ([x]).
    :param axis: The axis along which the vector components are aligned.
    :returns: The angle between two vectors x and y (rad).
    :raises ValueError: The shapes of x and y must be compatible.
    """
    x, y = np.asarray(x), np.asarray(y)
    check_shape_like(x, y)
    return np.arccos(np.sum(x * y, axis=axis) / np.sqrt(np.sum(x ** 2, axis=axis) * np.sum(y ** 2, axis=axis)))


def angle_d(x: array_like, x_d: array_like, y: array_like, y_d: array_like, axis=-1):
    """
    :param x: The first vectors (arb. units).
    :param x_d: The uncertainties of the first vectors ([x]).
    :param y: The second vectors ([x]).
    :param y_d: The uncertainties of the second vectors ([x]).
    :param axis: The axis along which the vector components are aligned.
    :returns: The angle between two vectors x and y (rad).
    :raises ValueError: The shapes of x and y must be compatible.
    """
    z = np.sum(x * y, axis=axis)
    n = np.sum(x ** 2, axis=axis) * np.sum(y ** 2, axis=axis)
    arg = z / np.sqrt(n)
    dx = np.sum(((n * y - z * x * np.sum(y ** 2, axis=axis)) / n ** 1.5 * x_d) ** 2, axis=axis)
    dy = np.sum(((n * x - z * y * np.sum(x ** 2, axis=axis)) / n ** 1.5 * y_d) ** 2, axis=axis)
    return np.sqrt((dx + dy) / (1 - arg ** 2))


def transform(t: array_like, vec: array_like, axis=-1) -> ndarray:
    """
    :param t: The transformation matrix which must hold t.shape[axis+1] == vec.shape[axis].
    :param vec: The vector to be transformed.
    :param axis: The axis along which the vector components of 'vec' are aligned and that is summed over.
    :returns: The transformed vector vec_new = t * vec.
    """
    ax = axis if axis != -1 else len(vec.shape) - 1
    t, vec = np.asarray(t), np.asarray(vec)
    # check_dimension(vec.shape[ax], ax + 1, t)
    vec = np.expand_dims(vec, ax)
    check_shape_like(t, vec)
    return np.sum(t * vec, axis=ax + 1)


def unit_vector(indexes: array_like, dim: int, axis: int = -1, dtype: type = float):
    """
    :param indexes: The indexes of the vectors that are filled with a 1.
    :param dim: The dimension of the unit vectors.
    :param axis: The axis with the dimension 'dim'.
    :param dtype: The data type of the unit vectors.
    :returns: Unit vectors in the specified 'indexes' with dimension 'dim' along the specified 'axis'.
    """
    index = np.asarray(indexes, dtype=int)
    return np.moveaxis(np.choose(np.expand_dims(index, axis=-1), np.identity(dim, dtype=dtype)), -1, axis)


def vector_to_diag_matrix(a, axis: int = -1):
    """
    :param a: The vector to be transformed into a diagonal matrix.
    :param axis: The axis of the vector elements to be aligned along the diagonal of the matrix.
    :returns: A diagonal matrix with one extra axis compared with 'a' at 'axis' + 1.
    """
    a = np.asarray(a)
    if axis < 0:
        axis += len(a.shape)
    shape = tuple(x1 if i > axis else x0 for i, (x0, x1) in enumerate(zip(a.shape + (1, ), (1, ) + a.shape)))
    b = np.broadcast_to(np.expand_dims(a, axis=axis + 1), shape)
    ident = np.identity(shape[axis], dtype=a.dtype)
    if axis > 0:
        ident = np.expand_dims(ident, axis=tuple(range(axis)))
    if len(a.shape) - 1 > axis:
        ident = np.expand_dims(ident, axis=tuple(range(axis + 2, len(b.shape))))
    return b * ident


def e_r(theta: array_like, phi: array_like, axis=-1) -> ndarray:
    """
    :param theta: The angle theta.
    :param phi: The angle phi.
    :param axis: The axis along which the vector components are aligned.
    :returns: The unit vector 'e_r'. Part of an orthonormal system defined by:
        e_x = e_r(pi/2, .)
        e_y = e_r(0, pi/2)
        e_z = e_r(0, 0)
    """
    x = np.expand_dims(np.sin(theta), axis=axis)
    y = np.expand_dims(np.cos(theta) * np.sin(phi), axis=axis)
    z = np.expand_dims(np.cos(theta) * np.cos(phi), axis=axis)
    return np.concatenate([x, y, z], axis=axis)


def e_theta(theta: array_like, phi: array_like, axis=-1) -> ndarray:
    """
    :param theta: The angle theta.
    :param phi: The angle phi.
    :param axis: The axis along which the vector components are aligned.
    :returns: The unit vector 'e_theta'. Part of an orthonormal system defined by:
        e_x = e_r(pi/2, .)
        e_y = e_r(0, pi/2)
        e_z = e_r(0, 0)
    """
    x = np.expand_dims(np.cos(theta) * np.ones_like(phi), axis=axis)
    y = np.expand_dims(-np.sin(theta) * np.sin(phi), axis=axis)
    z = np.expand_dims(-np.sin(theta) * np.cos(phi), axis=axis)
    return np.concatenate([x, y, z], axis=axis)


def e_phi(theta: array_like, phi: array_like, axis=-1) -> ndarray:
    """
    :param theta: The angle theta.
    :param phi: The angle phi.
    :param axis: The axis along which the vector components are aligned.
    :returns: The unit vector 'e_phi'. Part of an orthonormal system defined by:
        e_x = e_r(pi/2, .)
        e_y = e_r(0, pi/2)
        e_z = e_r(0, 0)
    """
    x = np.expand_dims(np.zeros_like(phi) * np.ones_like(theta), axis=axis)
    y = np.expand_dims(np.cos(phi) * np.ones_like(theta), axis=axis)
    z = np.expand_dims(-np.sin(phi) * np.ones_like(theta), axis=axis)
    return np.concatenate([x, y, z], axis=axis)


def orthonormal_rtp(r: array_like, axis=-1) -> (array_like, array_like, array_like):
    """
    :param r: The first vector of the orthonormal system. Does not have to be normalized.
    :param axis: The axis along which the vector components are aligned.
    :returns: The three orthonormal vectors 'e_r', 'e_theta' and 'e_phi'
     given the vector 'r' pointing in the 'e_r' direction.
    """
    r = np.asarray(r)
    check_dimension(3, axis, r)
    r_abs = np.expand_dims(absolute(r, axis=axis), axis=axis)
    e_r0 = r / r_abs
    theta = np.arcsin(np.take(e_r0, 0, axis=axis))
    valid = np.abs(np.abs(theta) - np.pi / 2.) > 1e-5
    phi = np.full_like(theta, 0.)
    if valid.size != 0:
        phi[valid] = np.arcsin(np.take(e_r0, 1, axis=axis)[valid] / np.cos(theta[valid]))
    return e_r0, e_theta(theta, phi, axis=axis), e_phi(theta, phi, axis=axis)


def orthonormal(r: array_like, axis=-1) -> (array_like, array_like, array_like):
    """
    :param r: The first vector of the orthonormal system. Does not have to be normalized.
    :param axis: The axis along which the vector components are aligned.
    :returns: Three orthonormal vectors, given the first vector 'r'.
    """
    r1 = np.asarray(r, dtype=float)
    check_dimension(3, axis, r1)
    r1 = np.moveaxis(r1, axis, -1)
    i_seed = np.argmin(np.abs(r1), axis=-1)
    w = unit_vector(i_seed, 3)
    r2 = np.cross(r1, w, axisa=-1, axisb=-1)
    r3 = np.cross(r1, r2, axisa=-1, axisb=-1)
    r1 /= np.expand_dims(absolute(r1, axis=-1), axis=-1)
    r2 /= np.expand_dims(absolute(r2, axis=-1), axis=-1)
    r3 /= np.expand_dims(absolute(r3, axis=-1), axis=-1)
    r1 = np.moveaxis(r1, -1, axis)
    r2 = np.moveaxis(r2, -1, axis)
    r3 = np.moveaxis(r3, -1, axis)
    return r1, r2, r3


def rotation_matrix(alpha: array_like, dr: array_iter):
    """
    :param alpha: The angle to rotate.
    :param dr: The vector to rotate about.
    :returns: The rotation matrix defined by the angle alpha and the vector dr.
    """
    alpha, dr = np.asarray(alpha), np.asarray(dr)
    dr /= absolute(dr)
    r = np.identity(3) * np.cos(alpha)
    r += np.array([[0., -dr[2], dr[1]], [dr[2], 0., -dr[0]], [-dr[1], dr[0], 0.]]) * np.sin(alpha)
    r += np.einsum('i,j->ij', dr, dr) * (1. - np.cos(alpha))
    return r


class Rotation:
    """
    An object specifying a rotation in 3d-space. The rotation is defined
    through an angle 'alpha' and a rotational axis 'dr' by the user.
    Additional instance attributes are the angle in degree 'alpha_deg' and the rotational matrix 'R'.
    """
    def __init__(self, alpha: scalar_like = 0., dr: array_iter = None):
        """
        :param alpha: The angle of the rotation (rad).
        :param dr: The rotational axis of the rotation.
        """
        self.alpha = alpha
        self.alpha_deg = self.alpha * 180. / np.pi
        if dr is None:
            dr = np.array([0., 0., 1.])
        self.dr = np.asarray(dr, dtype=float)
        check_shape((3,), self.dr, allow_scalar=False)
        self.dr /= absolute(self.dr)
        self.R = rotation_matrix(self.alpha, self.dr)


def rotation_to_vector(x: array_iter, y: array_iter):
    """
    :param x: The first vector.
    :param y: The second vector.
    :returns: A Rotation object which rotates the first vector onto the second.
    """
    _x, _y = np.asarray(x), np.asarray(y)
    check_shape((3, ), _x, _y, allow_scalar=False)
    _x /= absolute(_x)
    _y /= absolute(_y)
    diff = _y - _x
    if np.all(diff == 0.):
        return Rotation()
    elif np.all(diff == 2. * _y):
        _x, dr, _ = orthonormal(_x)
        return Rotation(alpha=np.pi, dr=dr)
    alpha = angle(x, y)
    dr = np.cross(x, y)
    return Rotation(alpha=alpha, dr=dr)


""" External program helpers """


def import_iso_shifts_tilda(db: str, iso_shifts: dict):
    """
    :param db: Location of a Tilda-compatible database.
    :param iso_shifts: A dictionary of isotope shifts
     with the structure {iso_str: {line_str: [val, stat_err, syst_err]}}.
    :returns: None. Writes the entries of the given dict to the specified Tilda-compatible database.
    """
    con = sqlite3.connect(db)
    cur = con.cursor()
    for iso, line_dict in iso_shifts.items():
        for line, val in line_dict.items():
            print(val)
            cur.execute('INSERT OR REPLACE INTO Combined (iso, parname, run) VALUES (?, ?, ?)',
                        (iso, 'shift', line))
            con.commit()
            cur.execute('UPDATE Combined SET config=?, final=?, '
                        'val = ?, statErr = ?, statErrForm=?, systErr = ?, systErrForm=? '
                        'WHERE iso = ? AND parname = ? AND run = ?',
                        ('[]', 0, val[0], val[1], 'err', val[2] if len(val) > 2 else 0., 0, iso, 'shift', line))
            con.commit()
    con.close()



# -*- coding: utf-8 -*-
"""
qspec.models._convolved
=======================

Convolution classes for lineshape models.
"""

import numpy as np

from qspec.tools import merge_intervals
from qspec.physics import source_energy_pdf
from qspec.models._base import Model
from qspec.models._spectrum import Gauss, Lorentz, GaussChi2

__all__ = ['CONVOLVE', 'Convolved', 'GaussConvolved', 'LorentzConvolved', 'GaussChi2Convolved']


CONVOLVE = ['None', 'Gauss', 'Lorentz', 'GaussChi2']


class Convolved(Model):
    def __init__(self, model_0, model_1):
        """
        A generic numerical convolution model.

        :param model_0: The first model to convolve.
        :param model_0: The second model to convolve.
        """
        super().__init__(model=model_0)
        self.type = 'Convolved'

        self.model_1 = model_1
        self.i_1 = self._index

        for (name, val, fix, link) in self.model_1.get_pars():
            self._add_arg('{}(conv)'.format(name), val, fix, link)

        self.j_1 = self._index

        self.precision = 8
        self.n = 2 ** self.precision + 1
        self.x_int = None

    def evaluate(self, x, *args, **kwargs):
        self.gen_x_int(*args)
        y = self.model.evaluate(np.expand_dims(x, axis=-1) - self.x_int, *args[:self.model.size], **kwargs) \
            * self.model_1.evaluate(self.x_int, *args[self.i_1:self.j_1])
        return np.trapz(y, dx=self.x_int[0, 1] - self.x_int[0, 0])

    def set_val(self, i, val, force=False):
        if force or isinstance(val, int) or isinstance(val, float):
            if self.model is None:
                self.vals[i] = val
            else:
                if self.i_1 <= i < self.j_1:
                    self.model_1.set_val(i - self.i_1, val, force=True)
                self.model.set_val(i, val, force=True)

    def min(self):
        return self.model.min() + self.model_1.min()

    def max(self):
        return self.model.max() + self.model_1.max()

    def intervals(self):
        return merge_intervals([[i[0] + self.model_1.min(), i[1] + self.model_1.max()]
                                for i in self.model.intervals()])

    @property
    def dx(self):
        return max([self.model.dx, self.model_1.dx])

    """ Preprocessing """

    def gen_x_int(self, *args):
        """
        Generates x-axis arrays for numerical integration.

        :param args: The function parameters.
        :returns: None.
        """
        temp_vals = [v for v in self.vals[:self.j_1]]
        self.set_vals(args[:self.j_1], force=True)
        dx = min([self.model.dx, self.model_1.dx])
        self.x_int = np.expand_dims(np.arange(self.model_1.min(), self.model_1.max() + 0.5 * dx, dx), axis=0)
        self.set_vals(temp_vals, force=True)


class GaussConvolved(Convolved):
    def __init__(self, model):
        """
        A convolution with a `Gauss` kernel.

        :param model: The model to convolve with a `Gauss` kernel.
        """
        super().__init__(model_0=model, model_1=Gauss())
        self.type = 'GaussConvolved'

    def evaluate(self, x, *args, **kwargs):  # Normalize the kernel function of the convolution to its integral.
        return super().evaluate(x, *args, **kwargs) / (np.sqrt(2 * np.pi) * args[self.i_1])


class LorentzConvolved(Convolved):
    def __init__(self, model):
        """
        A convolution with a `Lorentz` kernel.

        :param model: The model to convolve with a `Lorentz` kernel.
        """
        super().__init__(model_0=model, model_1=Lorentz())
        self.type = 'LorentzConvolved'

    def evaluate(self, x, *args, **kwargs):  # Normalize the kernel function of the convolution to its integral.
        return super().evaluate(x, *args, **kwargs) / (0.5 * np.pi * args[self.i_1])


class GaussChi2Convolved(Convolved):
    def __init__(self, model):
        """
        A convolution with a `GaussChi2` kernel.

        :param model: The model to convolve with a `GaussChi2` kernel.
        """
        super().__init__(model_0=model, model_1=GaussChi2())
        self.type = 'GaussChi2Convolved'

    def evaluate(self, x, *args, **kwargs):  # Normalize the kernel function of the convolution to its integral.
        return super().evaluate(x, *args, **kwargs) \
            * source_energy_pdf(0, 0, args[self.i_1], args[self.i_1 + 1], collinear=True)



# -*- coding: utf-8 -*-
"""
qspec.models._base
==================

Base classes for lineshape models.
"""

import numpy as np

from qspec.qtypes import *
from qspec.tools import merge_intervals

__all__ = ['Model', 'Empty', 'NPeak', 'Offset', 'Amplifier', 'Custom', 'YPars', 'Listed', 'Summed', 'Linked']


np_version = np.version.version.split('.')


def _is_scalar(val):
    return isinstance(val, bool) or np.issubdtype(type(val), np.integer) or np.issubdtype(type(val), np.floating)


def _is_unc(fix):
    if isinstance(fix, float):
        return True
    if not isinstance(fix, str):
        return False

    j, k = fix.find('('), fix.find(')')
    try:
        if j == -1 or k == -1:
            raise ValueError('Not a value with uncertainty.')
        _ = '{}({})'.format(float(fix[:j]), float(fix[j+1:k]))
    except ValueError:
        return False
    return True


def _val_fix_to_val(val, fix):
    if isinstance(fix, float):
        return val
    j = fix.find('(')
    return float(fix[:j])


def _fix_to_unc(fix):
    if isinstance(fix, float):
        return fix
    j, k = fix.find('('), fix.find(')')
    return float(fix[j+1:k])


def _args_ordered(args, order):
    return [args[i] for i in order]


def _poly(x, *args):
    return np.sum([args[n] * x ** n for n in range(len(args))], axis=0)


class Model:
    def __init__(self, model: 'Model' = None):
        """
        Base class for all models.

        :param model: A submodel whose parameters are adopted by this model.
        """
        self.model = model
        self.type = 'Model'

    def __call__(self, x: array_like, *args, **kwargs):
        return self.evaluate(x, *self.update_args(args), **kwargs)

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict) -> ndarray:
        """
        The main function of the model. This is executed when the model is called.
        Reimplement this function in subclasses.

        :param x: The input values $x$.
        :param args: The function parameters. Must have length `Model.size`.
        :param kwargs: Additional keyword arguments.
        :returns: The function values $y$ at the input values `x`.
        """
        pass

    def _add_arg(self, name, val, fix, link):
        """
        Add a new parameter to the model.

        :param name: The parameter name.
        :param val: The parameter value.
        :param fix: Whether the parameter is fixed.
        :param link: Whether the parameter is linked.
        """
        if name in self.names:
            raise ValueError('Parameter {} already exists.'.format(name))
        self.names.append(name)
        self.vals.append(val)
        self.fixes.append(fix)
        self.links.append(link)
        self.expressions.append('args[{}]'.format(self._index))

        self.p[name] = self._index

        self._index += 1
        self._size += 1

    @property
    def description(self):
        """
        A description of the model hierarchy.

        :returns: A `str` representing the model hierarchy.
        """
        label = ''
        super_model = self
        while super_model is not None:
            if isinstance(super_model, Listed):
                label += super_model.type + '[0].'
                super_model = super_model.models[0]
            else:
                label += super_model.type + '.'
                super_model = super_model.model
        return label[:-1]

    @property
    def size(self):
        """
        The number of parameters required by the model.
        """
        return self._size

    @property
    def dx(self):
        """
        A hint for an x-axis step size for a smooth display of the model.
        """
        return 0.1 if self.model is None else self.model.dx

    @property
    def error(self):
        """
        An error message if there is an issue with the model parameters (not implemented).
        """
        return self._error

    @property
    def model(self):
        """
        The submodel.
        """
        return self._model

    @model.setter
    def model(self, value):
        self._model = value
        if self._model is None:
            self.names, self.vals, self.fixes, self.links = [], [], [], []
            self.expressions = []
            self.p = {}
            self._index = 0
            self._size = 0
            self._error = ''
        else:
            self.names, self.vals, self.fixes, self.links = \
                self._model.names, self._model.vals, self._model.fixes, self._model.links
            self.expressions = self._model.expressions
            self.p = self._model.p
            self._index = len(self._model.names)
            self._size = len(self._model.names)
            self._error = self._model.error

    def get_pars(self):
        """
        :returns: A zip-iterator over (names, vals, fixes, links).
        """
        return zip(self.names, self.vals, self.fixes, self.links)

    def set_pars(self, pars, force=False):
        """
        Set all vals, fixes and links with one nested list.

        :param pars: A nested list of shape (self.size, 3).
        :param force: The `force` parameter of the `set_val`, `set_fix` and `set_link` functions.
        :returns: None.
        """
        for i, p in enumerate(pars):
            self.set_val(i, p[0], force=force)
            self.set_fix(i, p[1], force=force)
            self.set_link(i, p[2], force=force)

    def set_vals(self, vals, force=False):
        """
        Set all vals with one list.

        :param vals: A list of shape (`self.size`, ).
        :param force: The `force` parameter of the `set_val` function.
        :returns: None.
        """
        for i, val in enumerate(vals):
            self.set_val(i, val, force=force)

    def set_fixes(self, fixes, force=False):
        """
        Set all fixes with one list.

        :param fixes: A list of shape (`self.size`, ).
        :param force: The `force` parameter of the `set_fix` function.
        :returns: None.
        """
        for i, fix in enumerate(fixes):
            self.set_fix(i, fix, force=force)

    def set_links(self, links, force=False):
        """
        Set all links with one list.

        :param links: A list of shape (`self.size`, ).
        :param force: The `force` parameter of the `set_link` function.
        :returns: None.
        """
        for i, link in enumerate(links):
            self.set_link(i, link, force=force)

    def set_val(self, i, val, force=False):
        """
        Set a specific parameter value.

        :param i: The index of the parameter.
        :param val: The new parameter value.
        :param force: Force the parameter to take exactly the new value.
         If False, the parameter is converted to the correct format.
        :returns: None.
        """
        if force or _is_scalar(val):
            if self.model is None:
                self.vals[i] = val if force else float(val)
            else:
                self.model.set_val(i, val, force=False)  # Set val for all submodels
                # to ensure set_val is called for a ListedModel if one is part of the submodels.
                # This is only needed for the vals since these may be predicted by the specific model.
                # Everything else is handled by the top-model.
        else:
            raise ValueError(f'The parameter value {val} has the wrong format. Must be a floating.')

    def set_fix(self, i, fix, force=False):
        """
        Set a specific parameter fix state.

        :param i: The index of the parameter.
        :param fix: The new parameter value.
        :param force: Force the parameter to take exactly the new fix state.
         If False, the parameter is converted to the correct format.
        :returns: None.
        """
        if force:
            self.fixes[i] = fix
            return

        if _is_scalar(fix):
            if isinstance(fix, bool):
                fix = bool(fix)
            elif fix <= 0 or np.isinf(fix):
                fix = fix <= 0
            else:
                fix = float(fix)
            expr = 'args[{}]'.format(i)
        elif isinstance(fix, str):
            j, k = fix.find('('), fix.find(')')
            try:
                if j == -1 or k == -1:
                    raise ValueError('Not a value with uncertainty.')
                fix = '{}({})'.format(float(fix[:j]), float(fix[j+1:k]))
                expr = 'args[{}]'.format(i)
            except ValueError:
                _fix = fix
                for j, name in enumerate(self.names):
                    _fix = _fix.replace(name, 'eval(self.expressions[{}])'.format(j))
                expr = _fix
                try:
                    self._eval_zero_division(self.vals, expr)
                except (ValueError, TypeError, SyntaxError, NameError) as e:
                    print('Invalid expression for parameter \'{}\': {}. Got a {}.'.format(self.names[i], fix, repr(e)))
                    return
        elif isinstance(fix, (tuple, list, np.ndarray)):
            if len(fix) == 0:
                fix = [0, 1]
            elif len(fix) == 1:
                fix = [0, fix[0]]
            else:
                fix = list(fix[:2])
            expr = 'args[{}]'.format(i)
        else:
            raise ValueError(f'The parameter fix state {fix} has the wrong format.'
                             f' Must be a scalar, bool, str or iterable.')
        temp_expr = self.expressions[i]
        temp_fix = self.fixes[i]
        self.expressions[i] = compile(expr, '<string>', 'eval', optimize=2)  # Compile beforehand to save time.
        self.fixes[i] = fix
        try:
            self.update_args(self.vals)
        except RecursionError as e:
            print('Expression for {} with fix {} form a loop. Got a {}.'.format(self.names[i], fix, repr(e)))
            self.expressions[i] = temp_expr
            self.fixes[i] = temp_fix

    def set_link(self, i, link, force=False):
        """
        Set a specific parameter link state.

        :param i: The index of the parameter.
        :param link: The new parameter link state.
        :param force: Force the parameter to take exactly the new link state.
         If False, the parameter is converted to the correct format.
        :returns: None.
        """
        if force:
            self.links[i] = link
            return

        if _is_scalar(link):
            self.links[i] = bool(link)
        else:
            raise ValueError(f'The parameter link state {link} has the wrong format. Must be a scalar or bool.')

    def _eval_zero_division(self, args, expr):
        """
        Safely calculate parameter expressions including nans, infs and zero divisions.

        :param args: The parameters.
        :param expr: The parameter expression.
        :returns: The processed parameter or 0 in case of a nan, inf or zero division.
        """
        try:
            with np.errstate(divide='ignore', invalid='ignore'):
                ret = eval(expr, {}, {'self': self, 'args': args})
            if isinstance(ret, float):
                return 0. if np.isnan(ret) or np.isinf(ret) else ret
            ret[np.isnan(ret) + np.isinf(ret)] = 0.
            return ret
        except ZeroDivisionError:
            return np.zeros_like(args[0], dtype=float)

    def update_args(self, args):
        """
        :param args: The parameters.
        :returns: The parameters updated with the parameter expressions.
        """
        return tuple(self._eval_zero_division(args, expr) for expr in self.expressions)

    def update(self):
        """
        Updates `self.vals` with updated parameters, see self.update_args.

        :returns: None.
        """
        self.set_vals(self.update_args(self.vals), force=False)

    def min(self):
        """
        :returns: A hint for an x-axis minimum for a complete display of the model.
        """
        return -1. if self.model is None else self.model.min()

    def max(self):
        """
        :returns: A hint for an x-axis maximum for a complete display of the model.
        """
        return 1. if self.model is None else self.model.max()

    def intervals(self):
        """
        :returns: A list of x-axis intervals for a complete display of the model.
        """
        return [[self.min(), self.max()]] if self.model is None else self.model.intervals()

    def x(self):
        """
        :returns: An array of x values for a complete and smooth display of the model.
        """
        return np.concatenate([np.arange(i[0], i[1], self.dx, dtype=float) for i in self.intervals()], axis=0)

    def fit_prepare(self):
        """
        :returns: fixed, bounds. A list of bool values which parameters are not varied in a fit
         and a list of bounds for the fit parameters. See parameters `p0_fixed` and `bounds` of `qspec.curve_fit`.
        """
        bounds = (-np.inf, np.inf)
        fixed = [fix for fix in self.fixes]
        b_lower = []
        b_upper = []
        _bounds = False
        for i, fix in enumerate(self.fixes):
            if isinstance(fix, bool):
                b_lower.append(-np.inf)
                b_upper.append(np.inf)
            elif _is_unc(fix):
                b_lower.append(-np.inf)
                b_upper.append(np.inf)
                fixed[i] = False
            elif isinstance(fix, list):
                _bounds = True
                b_lower.append(fix[0])
                b_upper.append(fix[1])
                fixed[i] = False
            elif isinstance(fix, str):
                b_lower.append(-np.inf)
                b_upper.append(np.inf)
                fixed[i] = True
                _fix = fix
            else:
                raise TypeError('The type {} of the element {} in self.fixes is not supported.'
                                .format(type(fix), fix))
        if _bounds:
            bounds = (b_lower, b_upper)
        return fixed, bounds


class Empty(Model):
    def __init__(self):
        """
        An empty model, returning `numpy.zeros_like(x)`
        """
        super().__init__(model=None)
        self.type = 'Empty'

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        """
        An empty model, returning `numpy.zeros_like(x)`.

        :param x: The input values $x$.
        :param args: [].
        :param kwargs: Additional keyword arguments.
        :returns: The function values $y$ at the input values `x`.
        """
        return np.zeros_like(x)


class NPeak(Model):
    def __init__(self, model, n_peaks=1):
        """
        Evaluates the given `model` at the positions $x_i$ with scalings $p_i$ and `0 <= i < n_peaks`.

        :param model: A submodel whose parameters are adopted by this model.
        :param n_peaks: The number of times the submodel is copied.
        """
        super().__init__(model=model)
        self.type = 'NPeak'
        self.n_peaks = int(n_peaks)
        for n in range(self.n_peaks):
            self._add_arg('x{}'.format(n), 0., n == 0, False)
            self._add_arg('p{}'.format(n), 1., n == 0, False)

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        return np.sum([args[self.model.size + 2 * n + 1]
                       * self.model.evaluate(x - args[self.model.size + 2 * n], *args[:self.model.size])
                       for n in range(self.n_peaks)], axis=0)

    def min(self):
        min_center = min(self.vals[self.p['x{}'.format(n)]] for n in range(self.n_peaks))
        return min_center + self.model.min()

    def max(self):
        max_center = max(self.vals[self.p['x{}'.format(n)]] for n in range(self.n_peaks))
        return max_center + self.model.max()

    def intervals(self):
        return merge_intervals([[i[0] + self.vals[self.model.size + 2 * n],
                                 i[1] + self.vals[self.model.size + 2 * n]]
                                for i in self.model.intervals() for n in range(self.n_peaks)])


class Offset(Model):
    def __init__(self, model=None, x_cuts=None, offsets=None):
        """
        Cuts the x-axis and adds y-axis offsets to every segment.
        
        :param model: The submodel the offset will be added to. If None, the offset will be added to zero.
        :param x_cuts: x values where to cut the x-axis.
        :param offsets: A list of maximally considered polynomial orders for each slice.
         The list must have length len(x_cuts) + 1.
        """
        super().__init__(model=model)
        self.type = 'Offset'
        if x_cuts is None:
            x_cuts = []
        self.x_cuts = sorted(list(x_cuts))
        self.offsets = offsets
        if self.offsets is None:
            self.offsets = [0]
        if len(self.offsets) != len(self.x_cuts) + 1:
            raise ValueError('The parameter offset must be a list of size \'len(x_cuts) + 1\''
                             ' and contain the maximally considered polynomial order for each slice.')

        self.offset_map = []
        self.offset_masks = []
        self.update_on_call = True

        self.gen_offset_map()

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        """

        :param x:
        :param args:
        :param kwargs:
        :return:
        """
        if self._model is None:
            return self._offset(x, *args)
        return self.model.evaluate(x, *args[:self.model.size]) + self._offset(x, *args)

    def set_x_cuts(self, x_cuts):
        """
        Set the values where to cut the x-axis into intervals with individual offset parameters.

        :param x_cuts: A list of x values where to cut the x-axis.
        :returns: None.
        """
        x_cuts = list(x_cuts)
        if len(x_cuts) != len(self.x_cuts):
            raise ValueError('\'x_cuts\' must not change its size.')
        self.x_cuts = sorted(list(x_cuts))

    def _offset(self, x, *args):
        """
        :param x: The input values.
        :param args: The function parameters.
        :returns: The offset polynomial.
        """
        if self.update_on_call:
            self.gen_offset_masks(x)
        if not isinstance(args[0], float) and len(args[0].shape) > 1:
            ret = np.zeros((x.shape[0], args[0].shape[1]), dtype=float)
            for i, mask in enumerate(self.offset_masks):
                _mask = np.broadcast_to(mask, ret.shape)
                ret[_mask] = _poly(x[mask][:, None], *_args_ordered(args, self.offset_map[i])).flatten()
        else:
            ret = np.zeros_like(x)
            for i, mask in enumerate(self.offset_masks):
                ret[mask] = _poly(x[mask], *_args_ordered(args, self.offset_map[i]))
        return ret

    def gen_offset_map(self):
        """
        Generate the offset parameters and a map of the interval and polynomial order to the parameter index space.

        :returns: None.
        """
        self.offset_map = []
        for i, n in enumerate(self.offsets):
            self.offset_map.append([])
            for k in range(n + 1):
                self.offset_map[-1].append(self._index)
                self._add_arg('off{}e{}'.format(i, k), 0., False, False)

    """ Preprocessing """

    def gen_offset_masks(self, x):
        """
        Generate the array masks corresponding to the `x_cuts`.

        :param x: The input values.
        :returns: None.
        """
        self.offset_masks = []
        for x0, x1 in zip([np.min(x) - 1., ] + self.x_cuts, self.x_cuts + [np.max(x) + 1., ]):
            x_mean = 0.5 * (x0 + x1)
            self.offset_masks.append(np.abs(x - x_mean) < x1 - x_mean)

    def guess_offset(self, x, y):
        """
        Guess the first two polynomial orders for a given data set (const and linear).

        :param x: The input values.
        :param y: The y data.
        :returns: None.
        """
        for i, mask in enumerate(self.offset_masks):
            self.vals[self.p['off{}e0'.format(i)]] = 0.5 * (y[mask][0] + y[mask][-1])
            try:
                self.vals[self.p['off{}e1'.format(i)]] = (y[mask][-1] - y[mask][0]) / (x[mask][-1] - x[mask][0])
            except KeyError:
                return


class Amplifier(Model):
    def __init__(self, order=None):
        """
        A polynomial of order `order`.

        :param order: The maximum considered order of the polynomial.
        """
        super().__init__(model=None)
        self.type = 'Amplifier'
        if order is None:
            order = 1
        self.order = order
        for n in range(order + 1):
            self._add_arg('a{}'.format(n), 1. if n == 1 else 0., False, False)
        self._min = -10
        self._max = 10

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        self._min = np.min(x)
        self._max = np.max(x)
        return _poly(x, *args)

    @property
    def dx(self):
        return 1e-2

    def min(self):
        return self._min

    def max(self):
        return self._max


class Custom(Model):
    def __init__(self, model=None, parameters=None):
        """
        A model with custom parameters. Without a submodel, Custom returns the user-specified parameters as an array
        regardless of the input `x`. Otherwise, the submodel is called and the custom parameters
        can be connected to other parameters by the user.

        :param model: A submodel whose parameters are adopted by this model.
        :param parameters: A list of str, representing the names of the custom parameters.
        """
        super().__init__(model=model)
        self.type = 'Custom'
        if parameters is None:
            parameters = []
        self.parameters = parameters

        for p in self.parameters:
            self._add_arg(p, 0., False, False)

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        if self.model is None:
            return np.array(args, dtype=float)
        return self.model.evaluate(x, *args[:self.model.size], **kwargs)


class YPars(Model):
    def __init__(self, model):
        """
        Concatenates the *Prior* parameters of the submodel, that have uncertainties as `fix` states,
        with the y-axis array resulting from calling the submodel. This is used internally in `qspec.models.fit`.

        :param model: A submodel whose parameters are adopted by this model.
        """
        super().__init__(model=model)
        self.type = 'YPars'

        self.p_y = [i for i, fix in enumerate(self.model.fixes) if _is_unc(fix)]

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        return np.concatenate([self.model.evaluate(x, *args, **kwargs),
                               np.array([args[p_y] for p_y in self.p_y], dtype=float)], axis=0)


class Listed(Model):
    def __init__(self, models, labels=None):
        """
        An abstract class for models with multiple submodels.

        :param models: A list of submodels whose parameters are adopted by this model.
        :param labels: A list of labels with the same length as `models`.
         The labels are appended to the parameter names of each submodel.
        """
        super().__init__(model=None)
        self.type = 'Listed'

        self.models = models  # models is just a reference to a list defined somewhere else.
        self.labels = labels
        if self.labels is None:
            if len(self.models) == 1:
                self.labels = ['', ]
            else:
                self.labels = ['__{}'.format(i) for i in range(len(self.models))]

        self.slices = []
        self.model_map = []
        self.index_map = []
        for i, (model, label) in enumerate(zip(self.models, self.labels)):
            self.slices.append(slice(self._index, self._index + model.size, 1))
            for j, (name, val, fix, link) in enumerate(model.get_pars()):
                self.model_map.append(i)
                self.index_map.append(j)
                if isinstance(fix, str) and not _is_unc(fix):
                    for _name in model.names:
                        fix = fix.replace(_name, '{}{}'.format(_name, label))
                self._add_arg('{}{}'.format(name, label), val, fix, link)
        self.set_fixes(list(self.fixes))

    def set_val(self, i, val, force=False):
        super().set_val(i, val, force=force)
        if i < len(self.model_map):
            self.models[self.model_map[i]].set_val(self.index_map[i], self.vals[i], force=True)

    def set_fix(self, i, fix, force=False):
        super().set_fix(i, fix, force=force)
        if i < len(self.model_map):
            self.models[self.model_map[i]].set_fix(self.index_map[i], self.fixes[i], force=True)

    def set_link(self, i, link, force=False):
        super().set_link(i, link, force=force)
        if i < len(self.model_map):
            self.models[self.model_map[i]].set_link(self.index_map[i], self.links[i], force=True)

    def inherit_vals(self, force=False):
        """
        Inherit the parameter values of the submodels.

        :param force: The 'force' parameter of 'self.set_val'.
        :returns: None
        """
        self.set_vals([val for model in self.models for val in model.vals], force=force)

    def inherit_fixes(self, force=False):
        """
        Inherit the parameter fixes of the submodels.

        :param force: The 'force' parameter of 'self.set_fix'.
        :returns: None
        """
        self.set_fixes([fix for model in self.models for fix in model.fixes], force=force)

    def inherit_links(self, force=False):
        """
        Inherit the parameter links of the submodels.

        :param force: The 'force' parameter of 'self.set_link'.
        :returns: None
        """
        self.set_links([link for model in self.models for link in model.links], force=force)


class Summed(Listed):
    def __init__(self, models, labels=None):
        """
        A `Listed` model summing over all submodels with individual `center` and `int` parameters.

        :param models: A list of submodels whose parameters are adopted by this model.
        :param labels: A list of labels with the same length as `models`.
         The labels are appended to the parameter names of each submodel.
        """
        super().__init__(models, labels=labels)
        self.type = 'Summed'

        self.indices_add = []
        for n, (model, label) in enumerate(zip(self.models, self.labels)):
            self.indices_add.append([self._index, self._index + 1])
            self._add_arg('center{}'.format(label), 0., False, False)
            self._add_arg('int{}'.format(label), 1., False, False)

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        return np.sum([args[i[1]] * model.evaluate(x - args[i[0]], *args[_slice], **kwargs)
                       for model, _slice, i in zip(self.models, self.slices, self.indices_add)], axis=0)

    @property
    def dx(self):
        return min(model.dx for model in self.models)

    def min(self):
        return min(model.min() for model in self.models)

    def max(self):
        return max(model.max() for model in self.models)

    def intervals(self):
        return merge_intervals([[i[0] + self.vals[j[0]], i[1] + self.vals[j[0]]]
                                for model, j in zip(self.models, self.indices_add) for i in model.intervals()])


class Linked(Listed):
    def __init__(self, models):
        """
        A `Listed` model linking all `link=True` parameters of the submodels.

        :param models: A list of submodels whose parameters are adopted by this model.
        """
        super().__init__(models, labels=None)
        self.type = 'Linked'

        for i, (name, val, fix, link) in enumerate(self.get_pars()):
            if link and (not fix or isinstance(fix, list) or _is_unc(fix)):
                _name = name[:name.rfind('__')]
                for j, model in enumerate(self.models):
                    if j < self.model_map[i] and _name in model.names:
                        self.set_fix(i, '{}__{}'.format(_name, j))
                        break

    def evaluate(self, x: array_like, *args: array_iter, **kwargs: dict):
        return np.concatenate(tuple(model.evaluate(_x, *args[_slice], **kwargs)
                                    for model, _slice, _x in zip(self.models, self.slices, x)), axis=0)

    def set_fix(self, i, fix, force=False):
        super(Listed, self).set_fix(i, fix, force=force)

    def set_link(self, i, link, force=False):
        super(Listed, self).set_link(i, link, force=force)



# -*- coding: utf-8 -*-
"""
qspec.models._helper
====================

Helper functions for the models.
"""

import numpy as np

from qspec.qtypes import *
from qspec.models import _base, _convolved, _splitter, _spectrum


__all__ = ['gen_model', 'find_model', 'find_models']


def gen_model(ijj, shape: Union[str, type, _spectrum.Spectrum] = None, qi: bool = False, hf_mixing: bool = False,
              n_peaks: int = None, offsets: Union[int, list] = None, x_cuts: Union[int, float, list] = None,
              convolve: Union[str, type, _spectrum.Spectrum] = None) -> _base.Model:
    """
    Create a lineshape model to fit arbitrary atomic fluorescence spectra.

    :param ijj: The three or an Iterable of three quantum numbers $I$, $J_l$ and $J_u$.
     Must be a list with the format `[I, J_l, J_u]` or `[[I0, J0_l, J0_u], [I1, J1_l, J1_u], ...]`.
    :param shape: A str representation of or a `Spectrum` type.
    :param qi: Whether to use a quantum interference model.
    :param hf_mixing: Whether to use a hyperfine-induced mixing model. Not implemented for `qi=True`.
    :param n_peaks: The number of "peaks per resonance" or "lineshape duplicates".
    :param offsets: The orders of the offset polynomials of the separate x-axis intervals.
     Must be a list or a single value. In the former case `len(offsets) == len(x_cuts) + 1` must hold.
     If `offsets` is `None`, a single constant offset is assumed.
    :param x_cuts: The x values where to cut the x-axis. Must be a list or a single value.
     In the former case `len(offsets) == len(x_cuts) + 1` must hold.
     If `x_cuts` is `None`, the x-axis will not be cut.
    :param convolve: A str representation of or a `Convolved` type.
    :returns: The constructed lineshape model.
    """
    if qi and hf_mixing:
        raise NotImplementedError('QI with HF-mixing is not implemented yet.')
    ijj = np.asarray(ijj, float)
    if len(ijj.shape) == 1:
        ijj = np.expand_dims(ijj, axis=0)
    elif len(ijj.shape) != 2:
        raise ValueError('\'ijj\' must have shape (3, ) or (., 3) but has shape {}.'.format(ijj.shape))
    ijj = ijj.tolist()

    if shape is None or not shape:
        shape = _base.Empty
    elif isinstance(shape, str):
        if shape[0].islower():
            shape = shape[0].upper() + shape[1:]
            if shape not in _spectrum.SPECTRA:
                raise ValueError('The shape {} is not available. Choose one of {}'.format(shape, _spectrum.SPECTRA))
            shape = eval('_spectrum.{}'.format(shape), {'_spectrum': _spectrum})
    elif isinstance(shape, type) and issubclass(shape, _spectrum.Spectrum):
        pass
    else:
        raise ValueError('shape must be None, a str representation of or a Spectrum type.')

    if convolve is not None:
        if isinstance(convolve, str):
            if convolve[0].islower():
                convolve = convolve[0].upper() + convolve[1:]
                if convolve not in _convolved.CONVOLVE:
                    raise ValueError('The convolution {} is not available. Choose one of {}'
                                     .format(convolve, _convolved.CONVOLVE))
                convolve = eval('_convolved.{}Convolved'.format(convolve), {'_convolved': _convolved})
        elif isinstance(convolve, type) and issubclass(convolve, _convolved.Convolved):
            pass
        else:
            raise ValueError('convolve must be a str representation of or a Convolved type.')

    spl = _splitter.gen_splitter_model(qi=qi, hf_mixing=hf_mixing)
    npeaks_model = _splitter.SplitterSummed([spl(shape(), i, j_l, j_u, 'HF{}'.format(n))
                                             for n, (i, j_l, j_u) in enumerate(ijj)])

    if n_peaks is not None:
        npeaks_model = _base.NPeak(model=npeaks_model, n_peaks=n_peaks)
    if convolve is not None:
        npeaks_model = convolve(model=npeaks_model)

    offset_model = _base.Offset(model=npeaks_model, offsets=offsets, x_cuts=x_cuts)
    return offset_model
    

def find_model(model: _base.Model, sub_model: Union[_base.Model, type]):
    """
    :param model: The model to search in.
    :param sub_model: The submodel to find.
    :returns: The first submodel of type or with the same type as `sub_model`. If `model` already has the same type as
     `sub_model`, `model` will be returned. Returns `None` if `model` has no submodel `sub_model`.
    """
    model_type = sub_model
    if isinstance(sub_model, _base.Model):
        model_type = type(sub_model)
    if model is None:
        return None
    if isinstance(model, model_type):
        return model
    _model = model.model
    if _model is None and hasattr(model, 'models'):
        _model = model.models[0]
    return find_model(_model, sub_model)


def find_models(model: _base.Model, sub_model: Union[_base.Model, type], model_list: Iterable = None):
    """
    :param model: The model to search in.
    :param sub_model: The submodel to find.
    :param model_list: A list of models to append the found submodel to.
    :returns: This function returns a list of the first models of type or with the same type as `sub_model`
     for every branch in model.
     
    """
    model_type = sub_model
    if isinstance(sub_model, _base.Model):
        model_type = type(sub_model)
    if model_list is None:
        model_list = []
    if model is None:
        return model_list
    if isinstance(model, model_type):
        return model_list + [model, ]
    _model = [model.model]
    if _model is None and hasattr(model, 'models'):
        _model = [m for m in model.models]
    for m in _model:
        find_models(m, sub_model, model_list=model_list)
    return model_list



# -*- coding: utf-8 -*-
"""
qspec
=====

Modules for calculations surrounding Collinear Laser Spectroscopy.
"""

from qspec.algebra import *
from qspec.analyze import *
from qspec import models
from qspec.physics import *
from qspec import simulate
from qspec.stats import *
from qspec.tools import *



# -*- coding: utf-8 -*-
"""
qspec.models
============

Module for constructing lineshape models.
"""

from qspec.models import _base
from qspec.models import _convolved
from qspec.models import _spectrum
from qspec.models import _splitter
from qspec.models import _helper
from qspec.models import _fit
from qspec.models._base import *
from qspec.models._convolved import *
from qspec.models._spectrum import *
from qspec.models._splitter import *
from qspec.models._helper import *
from qspec.models._fit import *

__all__ = []
__all__.extend(_base.__all__)
__all__.extend(_convolved.__all__)
__all__.extend(_spectrum.__all__)
__all__.extend(_splitter.__all__)
__all__.extend(_helper.__all__)
__all__.extend(_fit.__all__)



# -*- coding: utf-8 -*-
"""
qspec.models._splitter
======================

Splitter classes for lineshape models.
"""

import os
from string import ascii_uppercase
import numpy as np

from qspec.tools import merge_intervals
from qspec.physics import get_f
from qspec.algebra import wigner_6j, a, b, c
from qspec.models._base import Model, Summed
from qspec.models._spectrum import LorentzQI

__all__ = ['gen_splitter_model', 'get_all_f', 'hf_coeff', 'hf_trans', 'hf_shift', 'hf_int', 'Splitter',
           'SplitterSummed', 'Hyperfine', 'HyperfineQI', 'HyperfineMixed']


def gen_splitter_model(qi: bool = False, hf_mixing: bool = False):
    """
    :param qi: Whether to consider QI effects.
    :param hf_mixing: Whether to consider HF mixing.
    :returns: The appropriate 'Splitter' model.
    """
    if qi and hf_mixing:
        pass
    elif qi:
        return HyperfineQI
    elif hf_mixing:
        return HyperfineMixed
    else:
        return Hyperfine
    raise ValueError('Specified splitter model not available.')


def get_all_f(i, j):
    """
    :param i: The nuclear spin.
    :param j: The total electron angular momentum.
    :returns: All possible F quantum numbers of an electronic finestructure state.
    """
    i, j = np.array(i, float).flatten(), np.array(j, float).flatten()
    return sorted(set(f + abs(_i - _j) for _i in i for _j in j for f in range(int(_i + _j - abs(_i - _j) + 1))))


def hf_coeff(i, j, f):
    """
    :param i: The nuclear spin.
    :param j: The total electron angular momentum.
    :param f: The total angular momentum.
    :returns: The tuple of hyperfine coefficients for A and B-factors of a given quantum state
    """
    # First and third order are taken from https://journals.aps.org/pra/abstract/10.1103/PhysRevA.103.032826.
    # Second order from https://link.springer.com/referencework/10.1007/978-0-387-26308-3.
    if i < 0.5 or j < 0.5:
        return tuple()

    # Magnetic dipole, the corresponding hyperfine constant is A = mu / (IJ) * <T_1>.
    k = f * (f + 1) - i * (i + 1) - j * (j + 1)
    co_a = 0.5 * k
    if i < 1 or j < 1:
        return co_a,

    # Electric quadrupole, the corresponding hyperfine constant is B = 2eQ * <T_2>.
    co_b = (0.75 * k * (k + 1) - j * (j + 1) * i * (i + 1)) / (2 * i * (2 * i - 1) * j * (2 * j - 1))
    if i < 1.5 or j < 1.5:
        return co_a, co_b

    # Magnetic octupole, the corresponding hyperfine constant is C = -Omega * <T_3>.
    co_c = k ** 3 + 4 * k ** 2 + 0.8 * k * (-3 * i * (i + 1) * j * (j + 1) + i * (i + 1) + j * (j + 1) + 3) \
        - 4 * i * (i + 1) * j * (j + 1)
    co_c /= i * (i - 1) * (2 * i - 1) * j * (j - 1) * (2 * j - 1)
    co_c *= 1.25
    if i < 2 or j < 2:
        return co_a, co_b, co_c

    return co_a, co_b, co_c  # Highest implemented order.


def hf_trans(i, j_l, j_u):
    """
    Calculate all allowed hyperfine transitions and their hyperfine coefficients.

    :returns: (f_l, f_u, coAl, coBl, coAu, coBu)
    """
    return [[(f_l, f_u), hf_coeff(i, j_l, f_l), hf_coeff(i, j_u, f_u)]
            for f_l in get_f(i, j_l) for f_u in get_f(i, j_u)
            if abs(f_u - f_l) < 1.1 and not f_u == f_l == 0]


def hf_shift(hyper_l, hyper_u, coeff_l, coeff_u):
    """
    :param hyper_l: The hyperfine structure constants of the lower state (Al, Bl, Cl, ...).
    :param hyper_u: The hyperfine structure constants of the upper state (Au, Bu, Cu, ...).
    :param coeff_l: The coefficients of the lower state to be multiplied by the constants (coAl, coBl, coCl, ...).
    :param coeff_u: The coefficients of the lower state to be multiplied by the constants (coAu, coBu, coCu, ...).
    :returns: The hyperfine structure shift of an optical transition.
    """
    return sum(const * coeff for const, coeff in zip(hyper_u, coeff_u)) \
        - sum(const * coeff for const, coeff in zip(hyper_l, coeff_l))


def hf_int(i, j_l, j_u, transitions):
    """
    :param i: The nuclear spin.
    :param j_l: The total lower state electron angular momentum.
    :param j_u: The total upper state electron angular momentum.
    :param transitions: A list of electronic transition properties. The first two properties of each entry should be
     f_l and f_u the total angular momenta of lower and upper state.
    :returns: The relative line intensities.
    """
    return [np.around((2 * f_u + 1) * (2 * f_l + 1) * (wigner_6j(j_l, f_l, i, f_u, j_u, 1, as_sympy=False) ** 2),
                      decimals=9) for (f_l, f_u), *r in transitions]


class Splitter(Model):
    def __init__(self, model, i, j_l, j_u, label=None):
        """
        The abstract base class for Hyperfine structure models.
        
        :param model: A submodel whose parameters are adopted by this model.
        :param i: The nuclear spin.
        :param j_l: The total lower state electron angular momentum.
        :param j_u: The total upper state electron angular momentum.
        :param label: A label for the `Splitter` (isotope / isomer / transition).
        """
        super().__init__(model=model)
        self.type = 'Splitter'

        self.i = i
        self.j_l = j_l
        self.j_u = j_u
        self.label = label

        self.racah_indices = []
        self.racah_intensities = []

    def racah(self):
        """
        Set the intensity values to the Racah intensities.

        :returns: None.
        """
        for i, intensity in zip(self.racah_indices, self.racah_intensities):
            self.vals[i] = intensity


class SplitterSummed(Summed):
    def __init__(self, splitter_models):
        """
        A `Summed` model of `Splitter` submodels.

        :param splitter_models: The `Splitter` submodels to sum over.
        """
        if any(not isinstance(model, Splitter) for model in splitter_models):
            raise TypeError('All models passed to \'SplitterSummed\' must have type \'Splitter\'.')
        super().__init__(splitter_models, labels=['({})'.format(i if model.label is None else model.label)
                                                  for i, model in enumerate(splitter_models)]
                         if len(splitter_models) > 1 else None)

    def racah(self):
        """
        Set the intensity values of all submodels to the Racah intensities.

        :returns: None.
        """
        i0 = 0
        for model in self.models:
            for i, intensity in zip(model.racah_indices, model.racah_intensities):
                self.set_val(i0 + i, intensity, force=True)
            i0 += model.size
        self.set_vals(self.vals, force=True)


class Hyperfine(Splitter):
    def __init__(self, model, i, j_l, j_u, label=None):
        """
        :param model: A submodel whose parameters are adopted by this model.
        :param i: The nuclear spin.
        :param j_l: The total lower state electron angular momentum.
        :param j_u: The total upper state electron angular momentum.
        :param label: A label for the `Splitter` (isotope / isomer / transition).
        """
        super().__init__(model, i, j_l, j_u, label=label)
        self.type = 'Hyperfine'

        self.transitions = hf_trans(self.i, self.j_l, self.j_u)
        self.racah_intensities = hf_int(self.i, self.j_l, self.j_u, self.transitions)

        self.n_l = len(self.transitions[0][1])
        self.n_u = len(self.transitions[0][2])
        for i in range(self.n_l):
            self._add_arg('{}l'.format(ascii_uppercase[i]), 0., False, False)
        for i in range(self.n_u):
            self._add_arg('{}u'.format(ascii_uppercase[i]), 0., False, False)

        for i in range(min([self.n_l, self.n_u])):
            fix = '{} / {}'.format('{}u'.format(ascii_uppercase[i]), '{}l'.format(ascii_uppercase[i]))
            self._add_arg('{}_ratio'.format(ascii_uppercase[i]), 0., fix, False)

        for i, (t, intensity) in enumerate(zip(self.transitions, self.racah_intensities)):
            self.racah_indices.append(self._index)
            self._add_arg('int({}, {})'.format(t[0][0], t[0][1]), intensity, i == 0, False)

    def evaluate(self, x, *args, **kwargs):
        const_l = tuple(args[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(args[self.model.size + self.n_l + i] for i in range(self.n_u))
        return np.sum([args[i] * self.model.evaluate(x - hf_shift(const_l, const_u, t[1], t[2]), *args, **kwargs)
                       for i, t in zip(self.racah_indices, self.transitions)], axis=0)

    def min(self):
        const_l = tuple(self.vals[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(self.vals[self.model.size + self.n_l + i] for i in range(self.n_u))
        return self.model.min() + min(hf_shift(const_l, const_u, t[1], t[2]) for t in self.transitions)

    def max(self):
        const_l = tuple(self.vals[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(self.vals[self.model.size + self.n_l + i] for i in range(self.n_u))
        return self.model.max() + max(hf_shift(const_l, const_u, t[1], t[2]) for t in self.transitions)

    def intervals(self):
        const_l = tuple(self.vals[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(self.vals[self.model.size + self.n_l + i] for i in range(self.n_u))
        shifts = [hf_shift(const_l, const_u, t[1], t[2]) for t in self.transitions]
        return merge_intervals([[self.model.min() + shift, self.model.max() + shift] for shift in shifts])

    def racah(self):
        for i, intensity in zip(self.racah_indices, self.racah_intensities):
            self.vals[i] = intensity


def load_qi(filepath):
    """
    :param filepath: The path to the saved model coefficients.
    :returns: The QI model coefficients A, B and C.
    """
    if not os.path.isfile(filepath):
        return None
    with open(os.path.join(filepath), 'r') as file:
        ret = eval(file.read())
    return ret['a'], ret['b'], ret['c']


class HyperfineQI(Splitter):
    def __init__(self, _, i, j_l, j_u, label=None, qi_path=None):
        """
        A perturbative quantum interference (QI) hyperfine-structure model
        based on https://doi.org/10.1103/PhysRevA.87.032504.
        
        :param _: Empty `model` parameter. This hyperfine-structure model always uses `Lorentz` lineshapes.
         To use other lineshapes, combine it with the `Convolve` model.
        :param i: The nuclear spin.
        :param j_l: The total lower state electron angular momentum.
        :param j_u: The total upper state electron angular momentum.
        :param label: A label for the `Splitter` (isotope / isomer / transition).
        :param qi_path: Specify a directory to save and load the calculated geometric parts of the
         dipole matrix elements for faster repeated calculations.
         These depend only on the quantum numbers `i`, `j_l` and `j_u`.
        """
        super().__init__(LorentzQI(), i, j_l, j_u, label)
        self.type = 'HyperfineQI'
        self.qi_path = qi_path
        if label is None:
            label = 'None'
        self.file = f'qi_{label}.txt'

        self.transitions = hf_trans(self.i, self.j_l, self.j_u)
        self.racah_intensities = [0., ]

        save = False
        if self.qi_path is None or not os.path.isfile(os.path.join(self.qi_path, self.file)):
            print(f'Calculating QI A-matrix of isotope {label} ... ')
            self.a_qi = [a(self.i, self.j_l, f_l, self.j_u, f_u, as_sympy=False)
                         for f_l in get_f(self.i, self.j_l) for f_u in get_f(self.i, self.j_u)
                         if abs(f_u - f_l) < 1.1 and not f_u == f_l == 0]
            print(f'Calculating QI B-matrix of isotope {label} ... ')
            self.b_qi = [b(self.i, self.j_l, f_l, self.j_u, f_u, as_sympy=False)
                         for f_l in get_f(self.i, self.j_l) for f_u in get_f(self.i, self.j_u)
                         if abs(f_u - f_l) < 1.1 and not f_u == f_l == 0]
            print(f'Calculating QI C-matrix of isotope {label} ... ')
            self.c_qi = [[c(self.i, self.j_l, f_l, self.j_u, f1_u, f2_u, as_sympy=False)
                          for i2, f2_u in enumerate(get_f(self.i, self.j_u)) if i2 > i1
                          if abs(f1_u - f_l) < 1.1 and abs(f2_u - f_l) < 1.1
                          and not f1_u == f_l == 0 and not f2_u == f_l == 0]
                         for f_l in get_f(self.i, self.j_l) for i1, f1_u in enumerate(get_f(self.i, self.j_u))]
            if qi_path is not None:
                save = True
        else:
            self.a_qi, self.b_qi, self.c_qi = load_qi(os.path.join(self.qi_path, self.file))
        if save:
            qi_dict = {'a': self.a_qi, 'b': self.b_qi, 'c': self.c_qi}
            with open(os.path.join(self.qi_path, self.file), 'w') as file:
                file.write(str(qi_dict))

        self.transitions_qi = [[[(f_l, f2_u), hf_coeff(self.i, self.j_l, f_l), hf_coeff(self.i, self.j_u, f2_u)]
                                for i2, f2_u in enumerate(get_f(self.i, self.j_u)) if i2 > i1
                                if abs(f1_u - f_l) < 1.1 and abs(f2_u - f_l) < 1.1
                                and not f1_u == f_l == 0 and not f2_u == f_l == 0]
                               for f_l in get_f(self.i, self.j_l) for i1, f1_u in enumerate(get_f(self.i, self.j_u))]

        self.n_l = len(self.transitions[0][1])
        self.n_u = len(self.transitions[0][2])
        for i in range(self.n_l):
            self._add_arg('{}l'.format(ascii_uppercase[i]), 0., False, False)
        for i in range(self.n_u):
            self._add_arg('{}u'.format(ascii_uppercase[i]), 0., False, False)

        for i in range(min([self.n_l, self.n_u])):
            fix = '{} / {}'.format('{}u'.format(ascii_uppercase[i]), '{}l'.format(ascii_uppercase[i]))
            self._add_arg('{}_ratio'.format(ascii_uppercase[i]), 0., fix, False)

        self.racah_indices.append(self._index)
        self._add_arg('geo', 0., False, False)

    def evaluate(self, x, *args, **kwargs):
        const_l = tuple(args[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(args[self.model.size + self.n_l + i] for i in range(self.n_u))
        return np.sum([(_a + _b * args[self.racah_indices[0]])
                       * self.model.evaluate(x - hf_shift(const_l, const_u, t[1], t[2]), *args, **kwargs)
                       + np.sum([_c * args[self.racah_indices[0]]
                                 * self.model.evaluate_qi(x - hf_shift(const_l, const_u, t[1], t[2]),
                                                          hf_shift(const_l, const_u, _t[1], _t[2])
                                                          - hf_shift(const_l, const_u, t[1], t[2]), *args, **kwargs)
                                 for _t, _c in zip(t_list, c_list)], axis=0)
                       for t, _a, _b, t_list, c_list in zip(
                self.transitions, self.a_qi, self.b_qi, self.transitions_qi, self.c_qi)], axis=0) / np.max(self.a_qi)

    def min(self):
        const_l = tuple(self.vals[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(self.vals[self.model.size + self.n_l + i] for i in range(self.n_u))
        return self.model.min() + min(hf_shift(const_l, const_u, t[1], t[2]) for t in self.transitions)

    def max(self):
        const_l = tuple(self.vals[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(self.vals[self.model.size + self.n_l + i] for i in range(self.n_u))
        return self.model.max() + max(hf_shift(const_l, const_u, t[1], t[2]) for t in self.transitions)

    def intervals(self):
        const_l = tuple(self.vals[self.model.size + i] for i in range(self.n_l))
        const_u = tuple(self.vals[self.model.size + self.n_l + i] for i in range(self.n_u))
        shifts = [hf_shift(const_l, const_u, t[1], t[2]) for t in self.transitions]
        return merge_intervals([[self.model.min() + shift, self.model.max() + shift] for shift in shifts])

    def racah(self):
        self.vals[self.racah_indices[0]] = 0.


class HyperfineMixed(Splitter):
    def __init__(self, model, i, j_l, j_u, label, config):
        """
        Hyperfine-mixing model based on https://doi.org/10.1103/PhysRevA.55.2728 [1].
        
        :param model: A submodel whose parameters are adopted by this model.
        :param i: The nuclear spin.
        :param j_l: The total lower state electron angular momentum.
        :param j_u: The total upper state electron angular momentum.
        :param label: A label for the `Splitter` (isotope / isomer / transition).
        :param config: The configuration of the hyperfine-induced mixing. This is a dictionary such as
         `{'enabled_l'=False, 'enabled_u'=False, 'Jl'=[0.5, ], 'Ju'=[0.5, ], 'Tl'=[[1.]], 'Tu'=[[1.]], 'fl'=[0., ], 'fu'=[0., ], 'mu'=0.}`,
         where 'enabled_l'/'enabled_u' decide if the corresponding lower/upper J mix, 'Jl'/'Ju' are lists of lower/upper
         state J, including `j_l` and `j_u`, 'Tl'/'Tu' are matrix representations of
         the electronic magnetic dipole operator, see [1], 'fl'/'fu' are lists of initial fine-structure energies
         (can be omitted) and 'mu' is the magnetic dipole moment of the nucleus.
        """
        super().__init__(model, i, j_l, j_u, label)
        self.type = 'HyperfineMixed'
        self.config = config
        self.states = ['l', 'u']
        to_mhz = 13074.70

        self.enabled = {s: self.config['enabled_{}'.format(s)] for s in self.states}

        self.J = {s: np.array(self.config['J{}'.format(s)]) for s in self.states}

        self.F = {s: get_all_f(self.i, self.J[s]) for s in self.states}

        self.T = {s: np.array(self.config['T{}'.format(s)], dtype=float) * to_mhz for s in self.states}

        self.M = 0. if self.i == 0 else np.sqrt((2 * self.i + 1) * (self.i + 1) / self.i) * self.config['mu']

        self.fs = {s: np.array(self.config.get('f{}'.format(s), np.zeros(self.J[s].shape)))
                   for s in self.states}

        self.mask_J = {s: [np.array([i for i, j in enumerate(self.config['J{}'.format(s)])
                                     if abs(self.i - j) - 0.1 < f < self.i + j + 0.1], dtype=int) for f in self.F[s]]
                       for s in self.states}

        self.W = {s: [np.array([[(-1) ** (self.i + self.J[s][i] + f)
                                 * wigner_6j(self.i, self.J[s][i], f, self.J[s][j], self.i, 1)
                                for j in self.mask_J[s][k]] for i in self.mask_J[s][k]], dtype=float)
                      * self.T[s][np.ix_(self.mask_J[s][k], self.mask_J[s][k])] * self.M
                      for k, f in enumerate(self.F[s])] for s in self.states}  # Eq. (2.5) in [1] without diagonal.

        self.transitions = [[(_m0, _m1), (self.J['l'][_m0], self.J['u'][_m1]), (f0, f1),
                             hf_coeff(self.i, self.J['l'][_m0], f0),
                             hf_coeff(self.i, self.J['u'][_m1], f1)]
                            for i, (m0, f0) in enumerate(zip(self.mask_J['l'], self.F['l']))
                            for j, (m1, f1) in enumerate(zip(self.mask_J['u'], self.F['u']))
                            for _m0 in m0 for _m1 in m1
                            if abs(f1 - f0) < 1.1 and abs(self.J['u'][_m1] - self.J['l'][_m0]) < 1.1]

        self.wt_map = {s: np.array([[(1 - 2 * int(s == 'l')) * int(t[0][int(s == 'u')] == _m
                                                                   and t[2][int(s == 'u')] == f)
                                     for m, f in zip(self.mask_J[s], self.F[s]) for _m in m]
                                    for t in self.transitions], dtype=float) for s in self.states}

        self.racah_intensities = [a(self.i, t[1][0], t[2][0], t[1][1], t[2][1]) for t in self.transitions]

        self.order = {s: [int(min((j // 0.5, self.i // 0.5))) for j in self.J[s]] for s in self.states}

        self.hf_args_map = [[] for _ in self.transitions]
        for s in self.states:
            if self.enabled[s] and self.i > 0:
                for i, j in enumerate(self.J[s]):
                    self._add_arg('FS_{}{}({})'.format(s, i, j), self.fs[s][i], i == 0, False)
            else:
                for i, j in enumerate(self.J[s]):
                    for k in range(self.order[s][i]):
                        for t, m in zip(self.transitions, self.hf_args_map):
                            if t[0][int(s == 'u')] == i:
                                m.append(self._index)
                        self._add_arg('{}_{}{}({})'.format(ascii_uppercase[k], s, i, j), 0., False, False)

        for i, (t, intensity) in enumerate(zip(self.transitions, self.racah_intensities)):
            self.racah_indices.append(self._index)
            self._add_arg('int{}([{}, {}] -> [{}, {}])'.format(i, t[1][0], t[2][0], t[1][1], t[2][1]),
                          intensity, i == 0, False)

    def x0(self, *args):
        """
        :param args: The function parameters.
        :returns: The peak positions.
        """
        x0 = np.zeros(len(self.transitions))
        for s in self.states:
            if self.enabled[s]:
                _x0 = [np.linalg.eigh(  # Eigenvalues and eigenvectors of Eq. (2.5) in [1].
                    np.diag([args[self.p['FS_{}{}({})'.format(s, _m, self.J[s][_m])]] for _m in m]) + w)
                       for m, w in zip(self.mask_J[s], self.W[s])]
                # Eigenvalues are returned in ascending order! Sort based on eigenvectors.
                # Invert order, e.g.,
                # a =               [a, b, c]
                # order =           [2, 0, 1]
                # desired a =       [b, c, a]
                # required order =  [1, 2, 0]
                orders = [np.argmax(np.abs(w[1]), axis=0) for w in _x0]
                orders = [np.array([j for i in range(order.size) for j, o in enumerate(order) if i == o])
                          for w, order in zip(_x0, orders)]
                _x0 = np.concatenate(tuple(w[0][order] for w, order in zip(_x0, orders)))
                _x0 = self.wt_map[s] @ _x0
            else:
                _x0 = np.array([sum(args[i] * coeff for i, coeff in zip(m, t[3 + int(s == 'u')]))
                                for m, t in zip(self.hf_args_map, self.transitions)])
                _x0 *= (1 - 2 * int(s == 'l'))
            x0 += _x0
        # print(f'[{", ".join([str(_x0) for _x0 in x0])}]')
        return x0

    def evaluate(self, x, *args, **kwargs):
        return np.sum([args[i] * self.model.evaluate(x - _x0, *args, **kwargs)
                       for i, _x0 in zip(self.racah_indices, self.x0(*args))], axis=0)

    def min(self):
        return self.model.min() + np.min(self.x0(*self.vals))

    def max(self):
        return self.model.max() + np.max(self.x0(*self.vals))

    def intervals(self):
        return merge_intervals([[self.model.min() + _x0, self.model.max() + _x0] for _x0 in self.x0(*self.vals)])

    def racah(self):
        for i, intensity in zip(self.racah_indices, self.racah_intensities):
            self.vals[i] = intensity
